from __future__ import annotations

import dataclasses
from dataclasses import dataclass
from typing import Any, Dict, List

import rasa.shared.utils.common
from rasa.shared.core.events import Event
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker


@dataclass
class Command:
    """A command that can be executed on a tracker."""

    @classmethod
    def type(cls) -> str:
        """Returns the type of the command."""
        raise NotImplementedError()

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        raise NotImplementedError()

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> Command:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        raise NotImplementedError()

    @staticmethod
    def command_from_json(data: Dict[str, Any]) -> "Command":
        """Converts a dictionary to a command object.

        First, resolves the command type and then converts the dictionary to
        the corresponding command object.

        Args:
            data: The dictionary to convert.

        Returns:
        The converted command object.
        """
        for cls in rasa.shared.utils.common.all_subclasses(Command):
            try:
                if data.get("command") == cls.command():
                    return cls.from_dict(data)
            except NotImplementedError:
                # we don't want to raise an error if the frame type is not
                # implemented, as this is ok to be raised by an abstract class
                pass
        else:
            raise ValueError(f"Unknown command type: {data}")

    def as_dict(self) -> Dict[str, Any]:
        """Converts the command to a dictionary.

        Returns:
            The converted dictionary.
        """
        data = dataclasses.asdict(self)
        data["command"] = self.command()
        return data

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        """Runs the command on the tracker.

        Args:
            tracker: The tracker to run the command on.
            all_flows: All flows in the assistant.
            original_tracker: The tracker before any command was executed.

        Returns:
            The events to apply to the tracker.
        """
        raise NotImplementedError()



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Generator, Optional, Text

from rasa.shared.core.flows.flow_step import FlowStep

if TYPE_CHECKING:
    from rasa.shared.core.flows.flow import Flow


@dataclass
class CallFlowStep(FlowStep):
    """Represents the configuration of an call flow step."""

    call: Text
    """The flow to be called."""
    called_flow_reference: Optional["Flow"] = None

    @classmethod
    def from_json(cls, flow_id: Text, data: Dict[Text, Any]) -> CallFlowStep:
        """Used to read flow steps from parsed YAML.

        Args:
            flow_id: The id of the flow that contains the step.
            data: The parsed YAML as a dictionary.

        Returns:
            The parsed flow step.
        """
        base = super().from_json(flow_id, data)
        return CallFlowStep(
            call=data.get("call", ""),
            **base.__dict__,
        )

    def as_json(self) -> Dict[Text, Any]:
        """Returns the flow step as a dictionary.

        Returns:
            The flow step as a dictionary.
        """
        dump = super().as_json()
        dump["call"] = self.call
        return dump

    def steps_in_tree(
        self, should_resolve_calls: bool = True
    ) -> Generator[FlowStep, None, None]:
        """Returns the steps in the tree of the flow step."""
        yield self

        if should_resolve_calls:
            if not self.called_flow_reference:
                raise ValueError("Call flow reference not set.")

            yield from self.called_flow_reference.steps_with_calls_resolved

        yield from self.next.steps_in_tree(should_resolve_calls)

    @property
    def default_id_postfix(self) -> str:
        """Returns the default id postfix of the flow step."""
        return f"call_{self.call}"



from __future__ import annotations

import re
import typing
from collections import defaultdict
from dataclasses import dataclass
from typing import List, Optional, Set, Text

from rasa.shared.constants import (
    RASA_DEFAULT_FLOW_PATTERN_PREFIX,
    RASA_PATTERN_HUMAN_HANDOFF,
    RASA_PATTERN_INTERNAL_ERROR,
)
from rasa.shared.core.flows.flow import Flow
from rasa.shared.core.flows.flow_step import (
    FlowStep,
)
from rasa.shared.core.flows.flow_step_links import (
    BranchingFlowStepLink,
    ElseFlowStepLink,
    IfFlowStepLink,
)
from rasa.shared.core.flows.flow_step_sequence import FlowStepSequence
from rasa.shared.core.flows.steps.call import CallFlowStep
from rasa.shared.core.flows.steps.collect import CollectInformationFlowStep
from rasa.shared.core.flows.steps.constants import CONTINUE_STEP_PREFIX, DEFAULT_STEPS
from rasa.shared.core.flows.steps.link import LinkFlowStep
from rasa.shared.core.flows.steps.set_slots import SetSlotsFlowStep
from rasa.shared.core.flows.utils import (
    get_duplicate_slot_persistence_config_error_message,
    get_invalid_slot_persistence_config_error_message,
    warn_deprecated_collect_step_config,
)
from rasa.shared.exceptions import RasaException

if typing.TYPE_CHECKING:
    from rasa.shared.core.flows.flows_list import FlowsList

FLOW_ID_REGEX = r"""^[a-zA-Z0-9_][a-zA-Z0-9_-]*?$"""


class UnreachableFlowStepException(RasaException):
    """Raised when a flow step is unreachable."""

    def __init__(self, step_id: str, flow_id: str) -> None:
        """Initializes the exception."""
        self.step_id = step_id
        self.flow_id = flow_id

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return (
            f"Step '{self.step_id}' in flow '{self.flow_id}' can not be reached "
            f"from the start step. Please make sure that all steps can be reached "
            f"from the start step, e.g. by "
            f"checking that another step points to this step."
        )


class MissingNextLinkException(RasaException):
    """Raised when a flow step is missing a next link."""

    def __init__(self, step_id: str, flow_id: str) -> None:
        """Initializes the exception."""
        self.step_id = step_id
        self.flow_id = flow_id

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return (
            f"Step '{self.step_id}' in flow '{self.flow_id}' is missing a `next`. "
            f"As a last step of a branch, it is required to have one. "
        )


class ReservedFlowStepIdException(RasaException):
    """Raised when a flow step is using a reserved id."""

    def __init__(self, step_id: str, flow_id: str) -> None:
        """Initializes the exception."""
        self.step_id = step_id
        self.flow_id = flow_id

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return (
            f"Step '{self.step_id}' in flow '{self.flow_id}' is using the reserved id "
            f"'{self.step_id}'. Please use a different id for your step."
        )


class DuplicatedStepIdException(RasaException):
    """Raised when a flow step is using the same id as another step."""

    def __init__(self, step_id: str, flow_id: str) -> None:
        """Initializes the exception."""
        self.step_id = step_id
        self.flow_id = flow_id

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return (
            f"Step '{self.step_id}' in flow '{self.flow_id}' is using the same id as "
            f"another step. Step ids must be unique across all steps of a flow. "
            f"Please use a different id for your step."
        )


class DuplicatedFlowIdException(RasaException):
    """Raised when a flow is using the same id as another flow."""

    def __init__(
        self, flow_id: str, first_file_path: str, second_file_path: str
    ) -> None:
        """Initializes the exception."""
        self.flow_id = flow_id
        self.first_file_path = first_file_path
        self.second_file_path = second_file_path

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        if self.first_file_path == self.second_file_path:
            return (
                f"Flow '{self.flow_id}' is used twice in `{self.first_file_path}`. "
                f"Please make sure flow IDs are unique across all files."
            )
        return (
            f"Flow '{self.flow_id}' is used in both "
            f"`{self.first_file_path}` and `{self.second_file_path}`. "
            f"Please make sure flow IDs are unique across all files."
        )


class MissingElseBranchException(RasaException):
    """Raised when a flow step is missing an else branch."""

    def __init__(self, step_id: str, flow_id: str) -> None:
        """Initializes the exception."""
        self.step_id = step_id
        self.flow_id = flow_id

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return (
            f"Step '{self.step_id}' in flow '{self.flow_id}' is missing an `else` "
            f"branch. If a steps `next` statement contains an `if` it always "
            f"also needs an `else` branch. Please add the missing `else` branch."
        )


class NoNextAllowedForLinkException(RasaException):
    """Raised when a flow step has a next link but is not allowed to have one."""

    def __init__(self, step_id: str, flow_id: str) -> None:
        """Initializes the exception."""
        self.step_id = step_id
        self.flow_id = flow_id

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return (
            f"Link step '{self.step_id}' in flow '{self.flow_id}' has a `next` but "
            f"as a link step is not allowed to have one."
        )


class ReferenceToPatternException(RasaException):
    """Raised when a flow step is referencing a pattern, which is not allowed."""

    def __init__(
        self, referenced_pattern: str, flow_id: str, step_id: str, call_step: bool
    ) -> None:
        """Initializes the exception."""
        self.step_id = step_id
        self.flow_id = flow_id
        self.referenced_pattern = referenced_pattern
        self.call_step = call_step

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        message = (
            f"Step '{self.step_id}' in flow '{self.flow_id}' is referencing a "
            f"pattern '{self.referenced_pattern}', which is not allowed. "
        )
        if self.call_step:
            return message + "Patterns can not be used as a target for a call step."
        else:
            return message + (
                "All patterns, except for 'pattern_human_handoff', can "
                "not be used as a target for a link step."
            )


class PatternReferencedPatternException(RasaException):
    """Raised when a pattern is referencing a pattern, which is not allowed."""

    def __init__(self, flow_id: str, step_id: str, call_step: bool) -> None:
        """Initializes the exception."""
        self.step_id = step_id
        self.flow_id = flow_id
        self.call_step = call_step

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        message = (
            f"Step '{self.step_id}' in pattern '{self.flow_id}' is referencing a "
            f"pattern which is not allowed. "
        )
        if self.call_step:
            return message + "Patterns can not use call steps to other patterns."
        else:
            return message + (
                "Patterns can not use link steps to other patterns. "
                "Exception: patterns can link to 'pattern_human_handoff'."
            )


class PatternReferencedFlowException(RasaException):
    """Raised when a pattern is referencing a flow, which is not allowed."""

    def __init__(self, flow_id: str, step_id: str, call_step: bool) -> None:
        """Initializes the exception."""
        self.step_id = step_id
        self.flow_id = flow_id
        self.call_step = call_step

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        message = (
            f"Step '{self.step_id}' in pattern '{self.flow_id}' is referencing a flow "
            f"which is not allowed. "
        )
        if self.call_step:
            return message + "Patterns can not use call steps."
        else:
            return message + (
                "'pattern_internal_error' can not use link steps to user flows."
            )


class NoLinkAllowedInCalledFlowException(RasaException):
    """Raised when a flow is called from another flow but is also using a link."""

    def __init__(self, step_id: str, flow_id: str, called_flow_id: str) -> None:
        """Initializes the exception."""
        self.step_id = step_id
        self.flow_id = flow_id
        self.called_flow_id = called_flow_id

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return (
            f"Flow '{self.flow_id}' is calling another flow (call step). "
            f"The flow that is getting called ('{self.called_flow_id}') is "
            f"using a link step, which is not allowed. "
            f"Either this flow can not be called or the link step in {self.step_id} "
            f"needs to be removed."
        )


class UnresolvedFlowException(RasaException):
    """Raised when a flow is called or linked from another flow but doesn't exist."""

    def __init__(self, flow_id: str, calling_flow_id: str, step_id: str) -> None:
        """Initializes the exception."""
        self.flow_id = flow_id
        self.calling_flow_id = calling_flow_id
        self.step_id = step_id

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return (
            f"Flow '{self.flow_id}' is called or linked from flow "
            f"'{self.calling_flow_id}' in step '{self.step_id}', "
            f"but it doesn't exist. "
            f"Please make sure that a flow with id '{self.flow_id}' exists."
        )


class UnresolvedFlowStepIdException(RasaException):
    """Raised when a flow step is referenced, but its id can not be resolved."""

    def __init__(
        self, step_id: str, flow_id: str, referenced_from_step_id: Optional[str]
    ) -> None:
        """Initializes the exception."""
        self.step_id = step_id
        self.flow_id = flow_id
        self.referenced_from_step_id = referenced_from_step_id

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        if self.referenced_from_step_id:
            exception_message = (
                f"Step with id '{self.step_id}' could not be resolved. "
                f"'Step '{self.referenced_from_step_id}' in flow '{self.flow_id}' "
                f"referenced this step but it does not exist. "
            )
        else:
            exception_message = (
                f"Step '{self.step_id}' in flow '{self.flow_id}' can not be resolved. "
            )

        return exception_message + (
            "Please make sure that the step is defined in the same flow."
        )


class EmptyStepSequenceException(RasaException):
    """Raised when an empty step sequence is encountered."""

    def __init__(self, flow_id: str, step_id: str) -> None:
        """Initializes the exception."""
        self.flow_id = flow_id
        self.step_id = step_id

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return (
            f"Encountered an empty step sequence in flow '{self.flow_id}' "
            f"and step '{self.step_id}'."
        )


class EmptyFlowException(RasaException):
    """Raised when a flow is completely empty."""

    def __init__(self, flow_id: str) -> None:
        """Initializes the exception."""
        self.flow_id = flow_id

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return f"Flow '{self.flow_id}' does not have any steps."


class DuplicateNLUTriggerException(RasaException):
    """Raised when multiple flows can be started by the same intent."""

    def __init__(self, intent: str, flow_names: List[str]) -> None:
        """Initializes the exception."""
        self.intent = intent
        self.flow_names = flow_names

    def __str__(self) -> Text:
        """Return a string representation of the exception."""
        return (
            f"The intent '{self.intent}' is used as 'nlu_trigger' "
            f"in multiple flows: {self.flow_names}. "
            f"An intent should just trigger one flow, not multiple."
        )


class SlotNamingException(RasaException):
    """Raised when a slot name to be collected does not adhere to naming convention."""

    def __init__(self, flow_id: str, step_id: str, slot_name: str) -> None:
        """Initializes the exception."""
        self.flow_id = flow_id
        self.step_id = step_id
        self.slot_name = slot_name

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return (
            f"For the flow '{self.flow_id}', collect step '{self.step_id}' "
            f"the slot name was set to : {self.slot_name}, while it has "
            f"to adhere to the following pattern: [a-zA-Z_][a-zA-Z0-9_-]*?."
        )


class FlowIdNamingException(RasaException):
    """Raised when a flow ID defined does not adhere to naming convention."""

    def __init__(self, flow_id: str) -> None:
        """Initializes the exception."""
        self.flow_id = flow_id

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return (
            f"The flow ID was set to : {self.flow_id}, while it has "
            f"to adhere to the following pattern: [a-zA-Z0-9_][a-zA-Z0-9_-]*?."
        )


class DuplicateSlotPersistConfigException(RasaException):
    """Raised when a slot persist configuration is duplicated."""

    def __init__(self, flow_id: str, collect_step: str) -> None:
        """Initializes the exception."""
        self.flow_id = flow_id
        self.collect_step = collect_step

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return get_duplicate_slot_persistence_config_error_message(
            self.flow_id, self.collect_step
        )


class InvalidPersistSlotsException(RasaException):
    """Raised when a slot persist configuration is duplicated."""

    def __init__(self, flow_id: str, invalid_slots: Set[str]) -> None:
        """Initializes the exception."""
        self.flow_id = flow_id
        self.invalid_slots = invalid_slots

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        return get_invalid_slot_persistence_config_error_message(
            self.flow_id, self.invalid_slots
        )


@dataclass
class ValidationResult:
    is_valid: bool
    invalid_slots: Set[str]


def validate_flow(flow: Flow) -> None:
    """Validates the flow configuration.

    This ensures that the flow semantically makes sense. E.g. it
    checks:
        - whether all next links point to existing steps
        - whether all steps can be reached from the start step
    """
    from rasa.cli.utils import is_skip_validation_flag_set

    validate_flow_not_empty(flow)
    validate_no_empty_step_sequences(flow)
    validate_all_steps_next_property(flow)
    validate_all_next_ids_are_available_steps(flow)
    validate_all_steps_can_be_reached(flow)
    validate_all_branches_have_an_else(flow)
    validate_not_using_builtin_ids(flow)
    validate_slot_names_to_be_collected(flow)
    validate_flow_id(flow)

    if is_skip_validation_flag_set():
        # we only want to run this validation if the --skip-validation flag is used
        # during training because Flow Validation exceptions are raised one by one
        # as opposed to all at once with the Validator class
        validate_slot_persistence_configuration(flow)


def validate_flow_not_empty(flow: Flow) -> None:
    """Validate that the flow is not empty."""
    if len(flow.steps) == 0:
        raise EmptyFlowException(flow.id)


def validate_no_empty_step_sequences(flow: Flow) -> None:
    """Validate that the flow does not have any empty step sequences."""
    for step in flow.steps:
        for link in step.next.links:
            if (
                isinstance(link, BranchingFlowStepLink)
                and isinstance(link.target_reference, FlowStepSequence)
                and len(link.target_reference.child_steps) == 0
            ):
                raise EmptyStepSequenceException(flow.id, step.id)


def validate_not_using_builtin_ids(flow: Flow) -> None:
    """Validates that the flow does not use any of the build in ids."""
    for step in flow.steps:
        if step.id in DEFAULT_STEPS or step.id.startswith(CONTINUE_STEP_PREFIX):
            raise ReservedFlowStepIdException(step.id, flow.id)


def validate_all_branches_have_an_else(flow: Flow) -> None:
    """Validates that all branches have an else link."""
    for step in flow.steps:
        links = step.next.links

        has_an_if = any(isinstance(link, IfFlowStepLink) for link in links)
        has_an_else = any(isinstance(link, ElseFlowStepLink) for link in links)

        if has_an_if and not has_an_else:
            raise MissingElseBranchException(step.id, flow.id)


def validate_all_steps_next_property(flow: Flow) -> None:
    """Validates that every step that must have a `next` has one."""
    for step in flow.steps:
        if isinstance(step, LinkFlowStep):
            # link steps can't have a next link!
            if not step.next.no_link_available():
                raise NoNextAllowedForLinkException(step.id, flow.id)
        elif step.next.no_link_available():
            # all other steps should have a next link
            raise MissingNextLinkException(step.id, flow.id)


def validate_all_next_ids_are_available_steps(flow: Flow) -> None:
    """Validates that all next links point to existing steps."""
    available_steps = {step.id for step in flow.steps} | DEFAULT_STEPS
    for step in flow.steps:
        for link in step.next.links:
            if link.target not in available_steps:
                raise UnresolvedFlowStepIdException(link.target, flow.id, step.id)


def validate_all_steps_can_be_reached(flow: Flow) -> None:
    """Validates that all steps can be reached from the start step."""

    def _reachable_steps(
        step: Optional[FlowStep], reached_steps: Set[Text]
    ) -> Set[Text]:
        """Validates that the given step can be reached from the start step."""
        if step is None or step.id in reached_steps:
            return reached_steps

        reached_steps.add(step.id)
        for link in step.next.links:
            reached_steps = _reachable_steps(
                flow.step_by_id(link.target), reached_steps
            )
        return reached_steps

    reached_steps = _reachable_steps(flow.first_step_in_flow(), set())

    for step in flow.steps:
        if step.id not in reached_steps:
            raise UnreachableFlowStepException(step.id, flow.id)


def validate_nlu_trigger(flows: List[Flow]) -> None:
    """Validates that an intent can just trigger one flow."""
    nlu_trigger_to_flows = defaultdict(list)

    for flow in flows:
        intents = flow.get_trigger_intents()
        for intent in intents:
            nlu_trigger_to_flows[intent].append(flow.name)

    for intent, flow_names in nlu_trigger_to_flows.items():
        if len(flow_names) > 1:
            raise DuplicateNLUTriggerException(intent, flow_names)


def validate_link_in_call_restriction(flows: "FlowsList") -> None:
    """Validates that a flow is not called from another flow and uses a link step."""

    def does_flow_use_link(flow_id: str) -> bool:
        if flow := flows.flow_by_id(flow_id):
            for step in flow.steps:
                if isinstance(step, LinkFlowStep):
                    return True
        return False

    for flow in flows.underlying_flows:
        for step in flow.steps:
            if isinstance(step, CallFlowStep) and does_flow_use_link(step.call):
                raise NoLinkAllowedInCalledFlowException(step.id, flow.id, step.call)


def validate_called_flows_exists(flows: "FlowsList") -> None:
    """Validates that all called flows exist."""
    for flow in flows.underlying_flows:
        for step in flow.steps:
            if not isinstance(step, CallFlowStep):
                continue

            if flows.flow_by_id(step.call) is None:
                raise UnresolvedFlowException(step.call, flow.id, step.id)


def validate_linked_flows_exists(flows: "FlowsList") -> None:
    """Validates that all linked flows exist."""
    for flow in flows.underlying_flows:
        for step in flow.steps:
            if not isinstance(step, LinkFlowStep):
                continue

            # It might be that the flows do not contain the default rasa patterns, but
            # only the user flows. Manually check for `pattern_human_handoff` as this
            # pattern can be linked to and it is part of the default patterns of rasa.
            if (
                flows.flow_by_id(step.link) is None
                and step.link != RASA_PATTERN_HUMAN_HANDOFF
            ):
                raise UnresolvedFlowException(step.link, flow.id, step.id)


def validate_patterns_are_not_called_or_linked(flows: "FlowsList") -> None:
    """Validates that patterns are never called or linked.

    Exception: pattern_human_handoff can be linked.
    """
    for flow in flows.underlying_flows:
        for step in flow.steps:
            if (
                isinstance(step, LinkFlowStep)
                and step.link.startswith(RASA_DEFAULT_FLOW_PATTERN_PREFIX)
                and step.link != RASA_PATTERN_HUMAN_HANDOFF
            ):
                raise ReferenceToPatternException(
                    step.link, flow.id, step.id, call_step=False
                )

            if isinstance(step, CallFlowStep) and step.call.startswith(
                RASA_DEFAULT_FLOW_PATTERN_PREFIX
            ):
                raise ReferenceToPatternException(
                    step.call, flow.id, step.id, call_step=True
                )


def validate_patterns_are_not_calling_or_linking_other_flows(
    flows: "FlowsList",
) -> None:
    """Validates that patterns do not contain call or link steps.

    Link steps to user flows are allowed for all patterns but 'pattern_internal_error'.
    Link steps to other patterns, except for 'pattern_human_handoff', are forbidden.
    """
    for flow in flows.underlying_flows:
        if not flow.is_rasa_default_flow:
            continue
        for step in flow.steps:
            if isinstance(step, LinkFlowStep):
                if step.link == RASA_PATTERN_HUMAN_HANDOFF:
                    # links to 'pattern_human_handoff' are allowed
                    continue
                if step.link.startswith(RASA_DEFAULT_FLOW_PATTERN_PREFIX):
                    # all other patterns are allowed to link to user flows, but not
                    # to other patterns
                    raise PatternReferencedPatternException(
                        flow.id, step.id, call_step=False
                    )
                if flow.id == RASA_PATTERN_INTERNAL_ERROR:
                    # 'pattern_internal_error' is not allowed to link at all
                    raise PatternReferencedFlowException(
                        flow.id, step.id, call_step=False
                    )
            if isinstance(step, CallFlowStep):
                if step.call.startswith(RASA_DEFAULT_FLOW_PATTERN_PREFIX):
                    raise PatternReferencedPatternException(
                        flow.id, step.id, call_step=True
                    )
                else:
                    raise PatternReferencedFlowException(
                        flow.id, step.id, call_step=True
                    )


def validate_step_ids_are_unique(flows: "FlowsList") -> None:
    """Validates that step ids are unique within a flow and any called flows."""
    for flow in flows.underlying_flows:
        used_ids: Set[str] = set()

        # check that the ids used in the flow are unique
        for step in flow.steps:
            if step.id in used_ids:
                raise DuplicatedStepIdException(step.id, flow.id)

            used_ids.add(step.id)


def validate_slot_names_to_be_collected(flow: Flow) -> None:
    """Validates that slot names to be collected comply with a specified regex."""
    slot_re = re.compile(r"""^[a-zA-Z_][a-zA-Z0-9_-]*?$""")
    for step in flow.steps:
        if isinstance(step, CollectInformationFlowStep):
            slot_name = step.collect
            if not slot_re.search(slot_name):
                raise SlotNamingException(flow.id, step.id, slot_name)


def validate_flow_id(flow: Flow) -> None:
    """Validates if the flow id comply with a specified regex.
    Flow IDs can start with an alphanumeric character or an underscore.
    Followed by zero or more alphanumeric characters, hyphens, or underscores.

    Args:
        flow: The flow to validate.

    Raises:
        FlowIdNamingException: If the flow id does not comply with the regex.
    """
    flow_re = re.compile(FLOW_ID_REGEX)
    if not flow_re.search(flow.id):
        raise FlowIdNamingException(flow.id)


def validate_slot_persistence_configuration(flow: Flow) -> None:
    """Validates that slot persistence configuration is valid.

    Only slots used in either a collect step or a set_slot step can be persisted
    and the configuration can either be set at the flow level or the collect step level,
    but not both.

    Args:
        flow: The flow to validate.

    Raises:
        DuplicateSlotPersistConfigException: If slot persist config is duplicated.
    """

    def _is_persist_slots_valid(
        persist_slots: List[str], flow_slots: Set[str]
    ) -> ValidationResult:
        """Validates that the slots that should be persisted are used in the flow."""
        invalid_slots = set(persist_slots) - flow_slots
        is_valid = False if invalid_slots else True

        return ValidationResult(is_valid, invalid_slots)

    flow_id = flow.id
    persist_slots = flow.persisted_slots
    has_flow_level_persistence = True if persist_slots else False
    flow_slots = set()

    for step in flow.steps_with_calls_resolved:
        if isinstance(step, SetSlotsFlowStep):
            flow_slots.update([slot["key"] for slot in step.slots])
        elif isinstance(step, CollectInformationFlowStep):
            flow_slots.add(step.collect)
            if not step.reset_after_flow_ends:
                collect_step = step.collect
                warn_deprecated_collect_step_config()
                if has_flow_level_persistence:
                    raise DuplicateSlotPersistConfigException(flow_id, collect_step)

    if has_flow_level_persistence:
        result = _is_persist_slots_valid(persist_slots, flow_slots)
        if not result.is_valid:
            raise InvalidPersistSlotsException(flow_id, result.invalid_slots)



from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable, Dict

import structlog
from keycloak.exceptions import KeycloakError
from requests.exceptions import ConnectionError, RequestException, Timeout

from rasa.shared.exceptions import RasaException
from rasa.shared.utils.cli import print_error, print_success
from rasa.studio.config import StudioConfig

structlogger = structlog.get_logger()


@dataclass
class StudioResult:
    message: str
    was_successful: bool

    @staticmethod
    def success(message: str) -> "StudioResult":
        return StudioResult(message, was_successful=True)

    @staticmethod
    def error(response: Dict[str, Any]) -> "StudioResult":
        """Create a StudioResult from a GraphQL error response.

        Factory will evaluate the response and return a StudioResult with the
        appropriate message and success status.
        """
        if isinstance(response.get("errors"), list):
            error_details = "; ".join(
                [error.get("message", "Unknown error") for error in response["errors"]]
            )
        else:
            error_details = "No detailed error information available."

        structlogger.warn(
            "studio.graphql_error", event_info=error_details, response=response
        )
        return StudioResult(error_details, was_successful=False)


def with_studio_error_handler(
    func: Callable[..., StudioResult],
) -> Callable[..., StudioResult]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            result = func(*args, **kwargs)
            if result.was_successful:
                print_success(result.message)
            else:
                print_error(result.message)
            return result
        except RasaException as e:
            return _handle_rasa_exception(e)
        except KeycloakError as e:
            return _handle_keycloak_error(e)
        except ConnectionError as e:
            return _handle_connection_error(e)
        except Timeout as e:
            return _handle_timeout_error(e)
        except RequestException as e:
            return _handle_request_exception(e)
        except Exception as e:
            return _handle_unexpected_error(e)

    return wrapper


def response_has_errors(response: Dict) -> bool:
    return (
        "errors" in response
        and isinstance(response["errors"], list)
        and len(response["errors"]) > 0
    )


def _handle_rasa_exception(e: RasaException) -> StudioResult:
    error_msg = "Rasa internal exception was raised while interacting with Studio."
    structlogger.error("studio.rasa_error", event_info=error_msg, exception=str(e))
    return StudioResult(message=str(e), was_successful=False)


def _handle_keycloak_error(e: KeycloakError) -> StudioResult:
    error_msg = (
        f"Unable to authenticate with Keycloak at "
        f"{StudioConfig.read_config().authentication_server_url} "
    )
    if e.response_code == 401:
        error_msg += "Please check if the credentials are correct."
    elif e.error_message.startswith("Can't connect to server"):
        error_msg += (
            "Please check if the server is running "
            "and the configured URL is correct. \n"
            "You may need to reconfigure Rasa Studio "
            "using 'rasa studio config'."
        )
    else:
        error_msg += f"Error message: {e.error_message}"
    structlogger.error("studio.keycloak_error", event_info=error_msg, exception=str(e))
    return StudioResult(error_msg, was_successful=False)


def _handle_connection_error(e: ConnectionError) -> StudioResult:
    studio_url = StudioConfig.read_config().studio_url
    error_msg = (
        f"Unable to reach Rasa Studio API at {studio_url} \n"
        "Please check if Studio is running and the configured URL is correct. \n"
        "You may need to reconfigure Rasa Studio using 'rasa studio config'."
    )
    structlogger.error("studio.graphql_error", event_info=error_msg, exception=str(e))
    return StudioResult(error_msg, was_successful=False)


def _handle_timeout_error(e: Timeout) -> StudioResult:
    error_msg = "The request timed out. Please try again later."
    structlogger.error("studio.graphql_timeout", event_info=error_msg, exception=str(e))
    return StudioResult(error_msg, was_successful=False)


def _handle_request_exception(e: RequestException) -> StudioResult:
    error_msg = f"An error occurred while communicating with the server: {e!s}"
    structlogger.error(
        "studio.graphql.request_exception", event_info=error_msg, exception=str(e)
    )
    return StudioResult(error_msg, was_successful=False)


def _handle_unexpected_error(e: Exception) -> StudioResult:
    error_msg = f"An unexpected error occurred: {e!s}"
    structlogger.exception(
        "studio.unexpected_error", event_info=error_msg, exception=str(e)
    )
    return StudioResult(error_msg, was_successful=False)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import structlog

from rasa.core.actions import action
from rasa.core.channels.channel import OutputChannel
from rasa.core.nlg.generator import NaturalLanguageGenerator
from rasa.dialogue_understanding.stack.frames import (
    BaseFlowStackFrame,
    PatternFlowStackFrame,
)
from rasa.shared.constants import RASA_DEFAULT_FLOW_PATTERN_PREFIX
from rasa.shared.core.constants import ACTION_CANCEL_FLOW
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import Event
from rasa.shared.core.flows.steps.constants import END_STEP
from rasa.shared.core.flows.steps.continuation import ContinueFlowStep
from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger()

FLOW_PATTERN_CANCEL = RASA_DEFAULT_FLOW_PATTERN_PREFIX + "cancel_flow"


@dataclass
class CancelPatternFlowStackFrame(PatternFlowStackFrame):
    """A pattern flow stack frame which cancels a flow.

    The frame contains the information about the stack frames that should
    be canceled."""

    flow_id: str = FLOW_PATTERN_CANCEL
    """The ID of the flow."""
    canceled_name: str = ""
    """The name of the flow that should be canceled."""
    canceled_frames: List[str] = field(default_factory=list)
    """The stack frames that should be canceled. These can be multiple
    frames since the user frame that is getting canceled might have
    created patterns that should be canceled as well."""

    @classmethod
    def type(cls) -> str:
        """Returns the type of the frame."""
        return FLOW_PATTERN_CANCEL

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> CancelPatternFlowStackFrame:
        """Creates a `DialogueStackFrame` from a dictionary.

        Args:
            data: The dictionary to create the `DialogueStackFrame` from.

        Returns:
            The created `DialogueStackFrame`.
        """
        return CancelPatternFlowStackFrame(
            frame_id=data["frame_id"],
            step_id=data["step_id"],
            canceled_name=data["canceled_name"],
            canceled_frames=data["canceled_frames"],
        )


class ActionCancelFlow(action.Action):
    """Action which cancels a flow from the stack."""

    def __init__(self) -> None:
        """Creates a `ActionCancelFlow`."""
        super().__init__()

    def name(self) -> str:
        """Return the flow name."""
        return ACTION_CANCEL_FLOW

    async def run(
        self,
        output_channel: OutputChannel,
        nlg: NaturalLanguageGenerator,
        tracker: DialogueStateTracker,
        domain: Domain,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> List[Event]:
        """Cancel the flow."""
        stack = tracker.stack
        if not (top := stack.top()):
            structlogger.warning("action.cancel_flow.no_active_flow")
            return []

        if not isinstance(top, CancelPatternFlowStackFrame):
            structlogger.warning("action.cancel_flow.no_cancel_frame", top=top)
            return []

        for canceled_frame_id in top.canceled_frames:
            for frame in stack.frames:
                if frame.frame_id == canceled_frame_id and isinstance(
                    frame, BaseFlowStackFrame
                ):
                    # Setting the stack frame to the end step so it is properly
                    # wrapped up by the flow policy
                    frame.step_id = ContinueFlowStep.continue_step_for_id(END_STEP)
                    break
            else:
                structlogger.warning(
                    "action.cancel_flow.frame_not_found",
                    dialogue_stack=stack,
                    frame_id=canceled_frame_id,
                )

        return tracker.create_stack_updated_events(stack)



import argparse
import asyncio
import sys
from pathlib import Path
from typing import List

import structlog

import rasa.cli.arguments.run
import rasa.cli.utils
import rasa.shared.data
import rasa.shared.utils.cli
import rasa.shared.utils.io
import rasa.utils.io
from rasa.cli import SubParsersAction
from rasa.cli.arguments.default_arguments import (
    add_endpoint_param,
    add_model_param,
    add_remote_storage_param,
)
from rasa.core.exceptions import AgentNotReady
from rasa.core.utils import AvailableEndpoints
from rasa.e2e_test.aggregate_test_stats_calculator import (
    AggregateTestStatsCalculator,
)
from rasa.e2e_test.constants import (
    DEFAULT_COVERAGE_OUTPUT_PATH,
    DEFAULT_E2E_INPUT_TESTS_PATH,
    DEFAULT_E2E_OUTPUT_TESTS_PATH,
    STATUS_FAILED,
    STATUS_PASSED,
)
from rasa.e2e_test.e2e_test_case import (
    KEY_STUB_CUSTOM_ACTIONS,
)
from rasa.e2e_test.e2e_test_coverage_report import (
    create_coverage_report,
    extract_tested_commands,
)
from rasa.e2e_test.e2e_test_runner import E2ETestRunner
from rasa.e2e_test.utils.io import (
    _save_coverage_report,
    _save_tested_commands_histogram,
    extract_test_case_from_path,
    print_test_result,
    read_test_cases,
    save_test_cases_to_yaml,
    split_into_passed_failed,
    write_test_results_to_file,
)
from rasa.e2e_test.utils.validation import validate_model_path
from rasa.exceptions import RasaException
from rasa.shared.constants import DEFAULT_ENDPOINTS_PATH, DEFAULT_MODELS_PATH
from rasa.utils.beta import ensure_beta_feature_is_enabled
from rasa.utils.endpoints import EndpointConfig

RASA_PRO_BETA_FINE_TUNING_RECIPE_ENV_VAR_NAME = "RASA_PRO_BETA_FINE_TUNING_RECIPE"

structlogger = structlog.get_logger()


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    """Add the e2e subparser to `rasa test`.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
    """
    for subparser in subparsers.choices.values():
        if subparser.prog == "rasa test":
            e2e_test_subparser = create_e2e_test_subparser(parents)

            for action in subparser._subparsers._actions:
                if action.choices is not None:
                    action.choices["e2e"] = e2e_test_subparser
                    return

    # If we get here, we couldn't hook the subparser to `rasa test`
    raise RasaException(
        "Hooking the e2e subparser to `rasa test` command "
        "could not be completed. Cannot run end-to-end testing."
    )


def create_e2e_test_subparser(
    parents: List[argparse.ArgumentParser],
) -> argparse.ArgumentParser:
    """Create e2e test subparser."""
    e2e_test_subparser = argparse.ArgumentParser(
        prog="rasa test e2e",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Runs end-to-end testing.",
    )

    e2e_test_subparser.set_defaults(func=execute_e2e_tests)

    add_e2e_test_arguments(e2e_test_subparser)
    add_model_param(e2e_test_subparser, add_positional_arg=False)
    add_endpoint_param(
        e2e_test_subparser,
        help_text="Configuration file for the model server and the connectors as a "
        "yml file.",
    )

    return e2e_test_subparser


def add_e2e_test_arguments(parser: argparse.ArgumentParser) -> None:
    """Arguments for running E2E tests directly using `rasa e2e`."""
    e2e_arguments = parser.add_argument_group("Testing Settings")
    e2e_arguments.add_argument(
        "path-to-test-cases",
        nargs="?",
        type=str,
        help="Input file or folder containing end-to-end test cases.",
        default=DEFAULT_E2E_INPUT_TESTS_PATH,
    )
    e2e_arguments.add_argument(
        "--fail-fast",
        action="store_true",
        help="Fail the test suite as soon as a unit test fails.",
    )

    parser.add_argument(
        "-o",
        "--e2e-results",
        action="store_const",
        const=DEFAULT_E2E_OUTPUT_TESTS_PATH,
        help="Results file containing end-to-end testing summary.",
    )

    add_remote_storage_param(parser)

    parser.add_argument(
        "--coverage-report",
        action="store_true",
        help="Generate a coverage report on flow paths and commands covered in e2e "
        "tests.",
    )

    parser.add_argument(
        "--coverage-output-path",
        default=DEFAULT_COVERAGE_OUTPUT_PATH,
        help="Directory where to save coverage report to.",
    )


def execute_e2e_tests(args: argparse.Namespace) -> None:
    """Run the end-to-end tests.

    Args:
        args: Commandline arguments.
    """
    if args.coverage_report:
        ensure_beta_feature_is_enabled(
            "LLM fine-tuning recipe",
            env_flag=RASA_PRO_BETA_FINE_TUNING_RECIPE_ENV_VAR_NAME,
        )

    args.endpoints = rasa.cli.utils.get_validated_path(
        args.endpoints, "endpoints", DEFAULT_ENDPOINTS_PATH, True
    )
    endpoints = AvailableEndpoints.get_instance(args.endpoints)

    # Ignore all endpoints apart from action server, model, nlu and nlg
    # to ensure InMemoryTrackerStore is being used instead of production
    # tracker store
    endpoints.tracker_store = None
    endpoints.lock_store = None
    endpoints.event_broker = None

    if endpoints.model is None:
        args.model = validate_model_path(args.model, "model", DEFAULT_MODELS_PATH)

    path_to_test_cases = getattr(
        args, "path-to-test-cases", DEFAULT_E2E_INPUT_TESTS_PATH
    )

    test_suite = read_test_cases(path_to_test_cases)

    if test_suite.stub_custom_actions:
        if not endpoints.action:
            endpoints.action = EndpointConfig()

        endpoints.action.kwargs[KEY_STUB_CUSTOM_ACTIONS] = (
            test_suite.stub_custom_actions
        )

    test_case_path, _ = extract_test_case_from_path(path_to_test_cases)

    try:
        test_runner = E2ETestRunner(
            remote_storage=args.remote_storage,
            model_path=args.model,
            model_server=endpoints.model,
            endpoints=endpoints,
            test_case_path=Path(test_case_path),
        )
    except AgentNotReady as error:
        structlogger.error(
            "rasa.e2e_test.execute_e2e_tests.agent_not_ready", message=error.message
        )
        sys.exit(1)

    results = asyncio.run(
        test_runner.run_tests(
            test_suite.test_cases,
            test_suite.fixtures,
            args.fail_fast,
            input_metadata=test_suite.metadata,
            coverage=args.coverage_report,
        )
    )

    passed, failed = split_into_passed_failed(results)

    if args.e2e_results is not None:
        results_path = Path(args.e2e_results)

        passed_file = rasa.cli.utils.get_e2e_results_file_name(
            results_path, STATUS_PASSED
        )
        write_test_results_to_file(passed, passed_file)

        failed_file = rasa.cli.utils.get_e2e_results_file_name(
            results_path, STATUS_FAILED
        )
        write_test_results_to_file(failed, failed_file)

    aggregate_stats_calculator = AggregateTestStatsCalculator(
        passed_results=passed, failed_results=failed, test_cases=test_suite.test_cases
    )
    accuracy_calculations = aggregate_stats_calculator.calculate()

    if args.coverage_report and test_runner.agent.processor:
        coverage_output_path = args.coverage_output_path
        rasa.shared.utils.io.create_directory(coverage_output_path)
        flows = asyncio.run(test_runner.agent.processor.get_flows())

        for results, status in [(passed, STATUS_PASSED), (failed, STATUS_FAILED)]:
            report = create_coverage_report(flows, results)
            _save_coverage_report(report, status, coverage_output_path)
            tested_commands = extract_tested_commands(results)
            _save_tested_commands_histogram(
                tested_commands, status, coverage_output_path
            )
            save_test_cases_to_yaml(results, coverage_output_path, status, test_suite)

        rasa.shared.utils.cli.print_info(
            f"Coverage data and result is written to {coverage_output_path}."
        )

    print_test_result(
        passed, failed, args.fail_fast, accuracy_calculations=accuracy_calculations
    )



from sanic import Sanic
from sanic.exceptions import ServerError
from sanic.handlers import ErrorHandler
from sanic.request import Request


# TODO: remove custom handler when upgrading to sanic >= 24
#       the underlying issue https://github.com/sanic-org/sanic/issues/2572
#       has been fixed in sanic 24
class IgnoreWSServerErrorHandler(ErrorHandler):
    @staticmethod
    def log(request: Request, exception: Exception) -> None:
        try:
            if (
                request.url.startswith("ws")
                and isinstance(exception, ServerError)
                and exception.args
                and (
                    exception.args[0]
                    == "Invalid response type None (need HTTPResponse)"
                )
            ):
                # in case we are in a websocket connection, we don't want to log the
                # the error, as this is a bug in sanic
                return
        except Exception:
            pass
        ErrorHandler.log(request, exception)  # type: ignore


def register_custom_sanic_error_handler(app: Sanic) -> None:
    app.error_handler = IgnoreWSServerErrorHandler()



from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Text, Union

import structlog

from rasa.dialogue_understanding_test.du_test_case import DialogueUnderstandingTestCase
from rasa.e2e_test.assertions import Assertion
from rasa.e2e_test.constants import (
    KEY_ASSERTION_ORDER_ENABLED,
    KEY_ASSERTIONS,
    KEY_BOT_INPUT,
    KEY_BOT_UTTERED,
    KEY_FIXTURES,
    KEY_METADATA,
    KEY_SLOT_NOT_SET,
    KEY_SLOT_SET,
    KEY_STEPS,
    KEY_STUB_CUSTOM_ACTIONS,
    KEY_TEST_CASE,
    KEY_TEST_CASES,
    KEY_USER_INPUT,
)
from rasa.e2e_test.stub_custom_action import StubCustomAction
from rasa.shared.constants import DOCS_BASE_URL
from rasa.shared.core.events import BotUttered, SlotSet, UserUttered
from rasa.shared.exceptions import RasaException

structlogger = structlog.get_logger(__name__)


@dataclass(frozen=True)
class Fixture:
    """Class for storing an input fixture."""

    name: Text
    slots_set: Dict[Text, Any]

    @staticmethod
    def from_dict(fixture_dict: Dict[Text, Any]) -> "Fixture":
        """Creates a fixture from a dictionary.

        Example:
            >>> Fixture.from_dict({"some_fixture": [{"slot_name": "slot_value"}]})
            Fixture(name="some_fixture", slots_set={"slot_name": "slot_value"})

        Args:
            fixture_dict: Dictionary containing the fixture.
        """
        return Fixture(
            name=next(iter(fixture_dict.keys())),
            slots_set={
                slot_name: slot_value
                for slots_list in fixture_dict.values()
                for slot_dict in slots_list
                for slot_name, slot_value in slot_dict.items()
            },
        )

    def as_dict(self) -> Dict[Text, Any]:
        """Returns the fixture as a dictionary."""
        return {
            self.name: [
                {slot_name: slot_value}
                for slot_name, slot_value in self.slots_set.items()
            ]
        }


@dataclass(frozen=True)
class TestStep:
    """Class for storing the input test step.

    Instances are frozen to make sure the underlying dictionary stays in sync
    with the text and actor attributes.
    """

    actor: Text
    text: Optional[Text] = None
    template: Optional[Text] = None
    line: Optional[int] = None
    slot_was_set: bool = False
    slot_was_not_set: bool = False
    _slot_instance: Optional[Union[Text, Dict[Text, Any]]] = None
    _underlying: Optional[Dict[Text, Any]] = None
    metadata_name: Optional[Text] = None
    assertions: Optional[List[Assertion]] = None
    assertion_order_enabled: bool = False

    @staticmethod
    def from_dict(test_step_dict: Dict[Text, Any]) -> "TestStep":
        """Creates a test step from a dictionary.

        Example:
            >>> TestStep.from_dict({"user": "hello"})
            TestStep(text="hello", actor="user")
            >>> TestStep.from_dict({"user": "hello", "metadata": "some_metadata"})
            TestStep(text="hello", actor="user", metadata_name="some_metadata")
            >>> TestStep.from_dict({"bot": "hello world"})
            TestStep(text="hello world", actor="bot")

        Args:
            test_step_dict: Dictionary containing the test step.
        """
        TestStep._validate_input_dict(test_step_dict)
        slot_instance = test_step_dict.get(KEY_SLOT_SET)
        if test_step_dict.get(KEY_SLOT_NOT_SET):
            slot_instance = test_step_dict.get(KEY_SLOT_NOT_SET)

        assertions = (
            [
                Assertion.create_typed_assertion(data)
                for data in test_step_dict.get(KEY_ASSERTIONS, [])
            ]
            if KEY_ASSERTIONS in test_step_dict
            else None
        )

        return TestStep(
            text=test_step_dict.get(
                KEY_USER_INPUT, test_step_dict.get(KEY_BOT_INPUT, "")
            ).strip()
            or None,
            template=test_step_dict.get(KEY_BOT_UTTERED),
            actor=KEY_USER_INPUT if KEY_USER_INPUT in test_step_dict else KEY_BOT_INPUT,
            line=test_step_dict.lc.line + 1 if hasattr(test_step_dict, "lc") else None,
            slot_was_set=bool(test_step_dict.get(KEY_SLOT_SET)),
            slot_was_not_set=bool(test_step_dict.get(KEY_SLOT_NOT_SET)),
            _slot_instance=slot_instance,
            _underlying=test_step_dict,
            metadata_name=test_step_dict.get(KEY_METADATA, ""),
            assertions=assertions,
            assertion_order_enabled=test_step_dict.get(
                KEY_ASSERTION_ORDER_ENABLED, False
            ),
        )

    @staticmethod
    def _validate_input_dict(test_step_dict: Dict[Text, Any]) -> None:
        if (
            KEY_USER_INPUT not in test_step_dict
            and KEY_BOT_INPUT not in test_step_dict
            and KEY_BOT_UTTERED not in test_step_dict
            and KEY_SLOT_SET not in test_step_dict
            and KEY_SLOT_NOT_SET not in test_step_dict
        ):
            raise RasaException(
                f"Test step is missing either the {KEY_USER_INPUT}, {KEY_BOT_INPUT}, "
                f"{KEY_SLOT_NOT_SET}, {KEY_SLOT_SET} "
                f"or {KEY_BOT_UTTERED} key: {test_step_dict}"
            )

        if (
            test_step_dict.get(KEY_SLOT_SET) is not None
            and test_step_dict.get(KEY_SLOT_NOT_SET) is not None
        ):
            raise RasaException(
                f"Test step has both {KEY_SLOT_SET} and {KEY_SLOT_NOT_SET} keys: "
                f"{test_step_dict}. You must only use one of the keys in a test step."
            )

        if KEY_USER_INPUT not in test_step_dict and KEY_ASSERTIONS in test_step_dict:
            raise RasaException(
                f"Test step with assertions must only be used with the "
                f"'{KEY_USER_INPUT}' key: {test_step_dict}"
            )

        if (
            KEY_USER_INPUT not in test_step_dict
            and KEY_ASSERTION_ORDER_ENABLED in test_step_dict
        ):
            raise RasaException(
                f"Test step with '{KEY_ASSERTION_ORDER_ENABLED}' key must "
                f"only be used with the '{KEY_USER_INPUT}' key: "
                f"{test_step_dict}"
            )

        if (
            KEY_ASSERTION_ORDER_ENABLED in test_step_dict
            and KEY_ASSERTIONS not in test_step_dict
        ):
            raise RasaException(
                f"You must specify the '{KEY_ASSERTIONS}' key in the user test step "
                f"where you are using '{KEY_ASSERTION_ORDER_ENABLED}' key: "
                f"{test_step_dict}"
            )

    def as_dict(self) -> Dict[Text, Any]:
        """Returns the underlying dictionary of the test step."""
        return self._underlying or {}

    def as_dict_yaml_format(self) -> Dict[Text, Any]:
        """Returns the test step as a dictionary in YAML format."""
        if not self._underlying:
            return {}

        result = self._underlying.copy()

        def _handle_slots(key: str) -> None:
            """Slots should be a list of strings or dicts."""
            if (
                self._underlying
                and key in self._underlying
                and isinstance(self._underlying[key], OrderedDict)
            ):
                result[key] = [
                    {key: value} for key, value in self._underlying[key].items()
                ]
            elif (
                self._underlying
                and key in self._underlying
                and isinstance(self._underlying[key], str)
            ):
                result[key] = [self._underlying[key]]

        _handle_slots(KEY_SLOT_SET)
        _handle_slots(KEY_SLOT_NOT_SET)

        return result

    def matches_event(self, other: Union[BotUttered, SlotSet, None]) -> bool:
        """Compares the test step with BotUttered or SlotSet event.

        Args:
            other: BotUttered or SlotSet event to compare with.

        Returns:
            `True` if the event and step match, `False` otherwise.
        """
        if other is None or self.actor != "bot":
            return False

        if isinstance(other, BotUttered):
            return self._do_utterances_match(other)
        elif isinstance(other, SlotSet):
            return self._do_slots_match(other)

        return False

    def _do_utterances_match(self, other: BotUttered) -> bool:
        if self.text:
            return other.text == self.text
        elif self.template:
            # we do set this utter_action metadata in the `ActionBotResponse`
            # action, so we can check if the template is the same
            return other.metadata.get("utter_action") == self.template

        return False

    def _do_slots_match(self, other: SlotSet) -> bool:
        if self._slot_instance:
            slot_name, slot_value = self.get_slot_name(), self.get_slot_value()
            if isinstance(self._slot_instance, dict):
                return other.key == slot_name and other.value == slot_value
            return other.key == slot_name

        return False

    def get_slot_name(self) -> Optional[Text]:
        """Returns the slot name if the test step is a slot action."""
        if isinstance(self._slot_instance, str):
            return self._slot_instance
        if isinstance(self._slot_instance, dict):
            return next(iter(self._slot_instance.keys()))
        return None

    def get_slot_value(self) -> Optional[Any]:
        """Returns the slot value if the test step is a slot action."""
        if isinstance(self._slot_instance, dict):
            return next(iter(self._slot_instance.values()))
        return None

    def is_slot_instance_dict(self) -> bool:
        """Asserts if the slot instance is a dictionary."""
        return isinstance(self._slot_instance, dict)


@dataclass
class ActualStepOutput:
    """Class for storing the events that are generated as a response from the bot."""

    actor: Text
    text: Optional[Text]
    events: List[Union[BotUttered, UserUttered, SlotSet]]
    bot_uttered_events: List[BotUttered]
    user_uttered_events: List[UserUttered]
    slot_set_events: List[SlotSet]

    @staticmethod
    def from_test_step(
        test_step: TestStep,
        events: List[Union[BotUttered, UserUttered, SlotSet]],
    ) -> "ActualStepOutput":
        """Creates a test step response from a test step and events.

        Example:
            >>> ActualStepOutput.from_test_step(TestStep(text="hello", actor="bot"))
            ActualStepOutput(text="hello", actor="bot")

        Args:
            test_step: Test step.
            events: Events generated by the bot.
        """
        bot_uttered_events, user_uttered_events, slot_set_events = [], [], []
        for event in events or []:
            if isinstance(event, BotUttered):
                bot_uttered_events.append(event)
            elif isinstance(event, UserUttered):
                user_uttered_events.append(event)
            elif isinstance(event, SlotSet):
                slot_set_events.append(event)

        return ActualStepOutput(
            text=test_step.text,
            actor=test_step.actor,
            events=events,
            bot_uttered_events=bot_uttered_events,
            user_uttered_events=user_uttered_events,
            slot_set_events=slot_set_events,
        )

    def remove_bot_uttered_event(self, bot_utter: BotUttered) -> None:
        """Removes bot_uttered_event from the ActualStepOutput."""
        if bot_utter in self.bot_uttered_events:
            self.bot_uttered_events.remove(bot_utter)
            self.events.remove(bot_utter)

    def remove_user_uttered_event(self, user_utter: UserUttered) -> None:
        """Removes user_uttered_event from the ActualStepOutput."""
        if user_utter in self.user_uttered_events:
            self.user_uttered_events.remove(user_utter)
            self.events.remove(user_utter)

    def remove_slot_set_event(self, slot_set: SlotSet) -> None:
        """Removes slot_set_event from the ActualStepOutput."""
        if slot_set in self.slot_set_events:
            self.slot_set_events.remove(slot_set)
            self.events.remove(slot_set)

    def get_user_uttered_event(self) -> Optional[UserUttered]:
        """Returns the user_uttered_event from the ActualStepOutput."""
        if self.user_uttered_events:
            # we assume that there is only one user_uttered_event
            # in the ActualStepOutput since the response is
            # only from one TestStep with one user utterance
            try:
                return self.user_uttered_events[0]
            except IndexError:
                structlogger.debug(
                    "e2e_test_case.get_user_uttered_event.no_user_uttered_event",
                    event_info=f"Could not find `UserUttered` event in the "
                    f"ActualStepOutput: {self}",
                )
                return None
        return None


@dataclass
class TestCase:
    """Class for storing the input test case."""

    name: Text
    steps: List[TestStep]
    file: Optional[Text] = None
    line: Optional[int] = None
    fixture_names: Optional[List[Text]] = None
    metadata_name: Optional[Text] = None

    @staticmethod
    def from_dict(
        input_test_case: Dict[Text, Any], file: Optional[Text] = None
    ) -> "TestCase":
        """Creates a test case from a dictionary.

        Example:
            >>> TestCase.from_dict({"test_case": "test", "steps": [{"user": "hello"}]})
            TestCase(name="test", steps=[TestStep(text="hello", actor="user")])

        Args:
            input_test_case: Dictionary containing the test case.
            file: File name of the test case.

        Returns:
            Test case object.
        """
        steps = []

        for step in input_test_case.get(KEY_STEPS, []):
            if KEY_SLOT_SET in step:
                for slot in step[KEY_SLOT_SET]:
                    step_copy = step.copy()
                    step_copy[KEY_SLOT_SET] = slot
                    steps.append(TestStep.from_dict(step_copy))
            elif KEY_SLOT_NOT_SET in step:
                for slot in step[KEY_SLOT_NOT_SET]:
                    step_copy = step.copy()
                    step_copy[KEY_SLOT_NOT_SET] = slot
                    steps.append(TestStep.from_dict(step_copy))
            else:
                steps.append(TestStep.from_dict(step))

        test_case = TestCase(
            name=input_test_case.get(KEY_TEST_CASE, "default"),
            steps=steps,
            file=file,
            line=(
                input_test_case.lc.line + 1 if hasattr(input_test_case, "lc") else None
            ),
            fixture_names=input_test_case.get(KEY_FIXTURES),
            metadata_name=input_test_case.get(KEY_METADATA),
        )
        test_case.validate()
        return test_case

    def validate(self) -> None:
        """Validates the test case.

        This method calls all validation methods required for the test case.
        """
        if self.uses_assertions():
            self.validate_duplicate_user_messages_metadata()

    def validate_duplicate_user_messages_metadata(self) -> None:
        """Validates that duplicate user messages use metadata correctly.

        Ensures that each duplicate user message uses unique metadata.

        Raises warnings if any issues are found.
        """
        docs_link = (
            f"{DOCS_BASE_URL}/testing/e2e-testing-assertions/assertions-how-to-guide/"
            "#how-to-handle-duplicate-user-text-messages-in-the-same-test-case"
        )
        no_metadata_event_info = (
            "Test case '{name}' has duplicate user steps with text '{text}', "
            "and user step at line {line} lacks metadata. When using "
            "duplicate user messages, metadata should be set on each step to ensure "
            f"correct processing. Please refer to the documentation: {docs_link}"
        )
        non_unique_metadata_event_info = (
            "Test case '{name}' has duplicate user steps with text '{text}', "
            "and user step at line {line} has duplicate metadata "
            "name '{metadata_name}'. Metadata names should be unique for each user "
            "step among duplicates. This may cause issues in processing "
            f"user messages. Please refer to the documentation: {docs_link}"
        )

        # Use dict[str, list] structure to group steps by user message text to easily
        # identify and validate instances with duplicate messages and their metadata.
        message_steps = defaultdict(list)

        # Collect user steps by text
        for step in self.steps:
            if step.actor == KEY_USER_INPUT and step.text:
                message_steps[step.text].append(step)

        # Check for duplicate messages
        for text, steps in message_steps.items():
            if len(steps) <= 1:
                continue

            metadata_names_used = set()
            for step in steps:
                if not step.metadata_name:
                    structlogger.warning(
                        "e2e_test_case.validate_duplicate_user_messages_metadata.no_metadata",
                        event_info=no_metadata_event_info.format(
                            name=self.name,
                            text=text,
                            line=step.line,
                        ),
                    )
                    break
                elif step.metadata_name in metadata_names_used:
                    structlogger.warning(
                        "e2e_test_case.validate_duplicate_user_messages_metadata.non_unique_metadata",
                        event_info=non_unique_metadata_event_info.format(
                            name=self.name,
                            text=text,
                            line=step.line,
                            metadata_name=step.metadata_name,
                        ),
                    )
                    break
                else:
                    metadata_names_used.add(step.metadata_name)

    def as_dict(self) -> Dict[Text, Any]:
        """Returns the test case as a dictionary."""
        result = {
            KEY_TEST_CASE: self.name,
            KEY_STEPS: [step.as_dict_yaml_format() for step in self.steps],
        }
        if self.fixture_names:
            result[KEY_FIXTURES] = self.fixture_names
        if self.metadata_name:
            result[KEY_METADATA] = self.metadata_name
        return result

    def file_with_line(self) -> Text:
        """Returns the file name and line number of the test case."""
        if not self.file:
            return ""

        line = str(self.line) if self.line is not None else ""
        return f"{self.file}:{line}"

    def uses_assertions(self) -> bool:
        """Checks if the test case uses assertions."""
        try:
            next(step for step in self.steps if step.assertions is not None)
        except StopIteration:
            return False

        return True


@dataclass(frozen=True)
class Metadata:
    """Class for storing an input metadata."""

    name: Text
    metadata: Dict[Text, Any]

    @staticmethod
    def from_dict(metadata_dict: Dict[Text, Any]) -> "Metadata":
        """Creates a metadata from a dictionary.

        Example:
            >>> Metadata.from_dict({"some_metadata": {"room": "test_room"}})
            Metadata(name="some_metadata", metadata={"room": "test_room"})

        Args:
            metadata_dict: Dictionary containing the metadata.
        """
        return Metadata(
            name=next(iter(metadata_dict.keys())),
            metadata={
                metadata_name: metadata_value
                for metadata in metadata_dict.values()
                for metadata_name, metadata_value in metadata.items()
            },
        )

    def as_dict(self) -> Dict[Text, Any]:
        """Returns the metadata as a dictionary."""
        return {self.name: self.metadata}


@dataclass(frozen=True)
class TestSuite:
    """Class for representing all top level test suite keys."""

    test_cases: List[Union[TestCase, DialogueUnderstandingTestCase]]
    fixtures: List[Fixture]
    metadata: List[Metadata]
    stub_custom_actions: Dict[Text, StubCustomAction]

    def as_dict(self) -> Dict[Text, Any]:
        """Returns the test suite as a dictionary."""
        return {
            KEY_FIXTURES: [fixture.as_dict() for fixture in self.fixtures],
            KEY_METADATA: [metadata.as_dict() for metadata in self.metadata],
            KEY_STUB_CUSTOM_ACTIONS: {
                key: value.as_dict() for key, value in self.stub_custom_actions.items()
            },
            KEY_TEST_CASES: [test_case.as_dict() for test_case in self.test_cases],
        }



import os
from functools import cache
from http import HTTPStatus
from typing import Any, Dict, Optional

import jwt
import requests
import structlog

from rasa.shared.exceptions import RasaException

structlogger = structlog.get_logger()


AUTH_URL = os.getenv("KEYCLOAK_URL", "http://localhost:8081/auth")

AUTH_REALM = os.getenv("KEYCLOAK_REALM", "rasa-studio")


class UserToServiceAuthenticationError(RasaException):
    """Raised when the user authentication fails."""

    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return f"{self.__class__.__name__}: {self.message}"


@cache
def get_public_key_from_keycloak() -> Optional[str]:
    """Fetch the public key from the keycloak server."""
    realm_url = f"{AUTH_URL}/realms/{AUTH_REALM}"

    try:
        response = requests.get(realm_url)
    except requests.RequestException as error:
        structlogger.error("model_api.auth.keycloak_request_failed", error=str(error))
        return None

    if response.status_code != HTTPStatus.OK:
        structlogger.error(
            "model_api.auth.keycloak_public_key_fetch_failed",
            status_code=response.status_code,
            response=response.text,
        )
        return None

    public_key = response.json().get("public_key")

    if public_key is None:
        structlogger.error(
            "model_runner.keycloak_public_key_not_found",
            response=response.text,
        )
        return None

    public_key = f"-----BEGIN PUBLIC KEY-----\n{public_key}\n-----END PUBLIC KEY-----"
    return public_key


def authenticate_user_to_service(token: str) -> Dict[str, Any]:
    """Authenticate the user to the model service."""
    if not token:
        structlogger.debug("model_api.auth.no_token_provided")
        raise UserToServiceAuthenticationError("No token provided.")

    public_key = get_public_key_from_keycloak()

    if public_key is None:
        raise UserToServiceAuthenticationError(
            "Failed to fetch public key from keycloak."
        )

    try:
        return jwt.decode(
            token,
            public_key,
            algorithms=["RS256", "HS256", "HS512", "ES256"],
            audience="account",
        )
    except jwt.InvalidKeyError as error:
        structlogger.info("model_api.auth.invalid_jwt_key", error=str(error))
        raise UserToServiceAuthenticationError("Invalid JWT key.")
    except jwt.InvalidTokenError as error:
        structlogger.info("model_api.auth.invalid_jwt_token", error=str(error))
        raise UserToServiceAuthenticationError("Invalid JWT token.") from error



import os
import ssl
from types import ModuleType
from typing import Any, Dict, List, Optional, Text, Union

import aiohttp
import structlog
from aiohttp.client_exceptions import ContentTypeError
from sanic.request import Request

from rasa.core.constants import DEFAULT_REQUEST_TIMEOUT
from rasa.shared.exceptions import FileNotFoundException
from rasa.shared.utils.yaml import read_config_file

structlogger = structlog.get_logger()


def read_endpoint_config(
    filename: Text, endpoint_type: Text
) -> Optional["EndpointConfig"]:
    """Read an endpoint configuration file from disk and extract one config."""
    if not filename:
        return None

    try:
        content = read_config_file(filename)

        if content.get(endpoint_type) is None:
            return None

        return EndpointConfig.from_dict(content[endpoint_type])
    except FileNotFoundError:
        structlogger.error(
            "endpoint.read.failed_no_such_file",
            filename=os.path.abspath(filename),
            event_info=(
                "Failed to read endpoint configuration file - "
                "the file was not found."
            ),
        )
        return None


def read_property_config_from_endpoints_file(
    filename: str, property_name: str
) -> Optional[Union[Dict[str, Any], List]]:
    """Read a property from an endpoint configuration file."""
    if not filename:
        return None

    try:
        content = read_config_file(filename)

        if content.get(property_name) is None:
            return None

        return content[property_name]
    except FileNotFoundError:
        structlogger.error(
            "endpoint.read.failed_no_such_file",
            filename=os.path.abspath(filename),
            event_info=(
                "Failed to read endpoint configuration file - "
                "the file was not found."
            ),
        )
        return None


def concat_url(base: Text, subpath: Optional[Text]) -> Text:
    """Append a subpath to a base url.

    Strips leading slashes from the subpath if necessary. This behaves
    differently than `urlparse.urljoin` and will not treat the subpath
    as a base url if it starts with `/` but will always append it to the
    `base`.

    Args:
        base: Base URL.
        subpath: Optional path to append to the base URL.

    Returns:
        Concatenated URL with base and subpath.
    """
    if not subpath:
        if base.endswith("/"):
            structlogger.debug(
                "endpoint.concat_url.trailing_slash",
                url=base,
                event_info=(
                    "The URL has a trailing slash. Please make sure the "
                    "target server supports trailing slashes for this endpoint."
                ),
            )
        return base

    url = base
    if not base.endswith("/"):
        url += "/"
    if subpath.startswith("/"):
        subpath = subpath[1:]
    return url + subpath


class EndpointConfig:
    """Configuration for an external HTTP endpoint."""

    def __init__(
        self,
        url: Optional[Text] = None,
        params: Optional[Dict[Text, Any]] = None,
        headers: Optional[Dict[Text, Any]] = None,
        basic_auth: Optional[Dict[Text, Text]] = None,
        token: Optional[Text] = None,
        token_name: Text = "token",
        cafile: Optional[Text] = None,
        actions_module: Optional[Union[Text, ModuleType]] = None,
        **kwargs: Any,
    ) -> None:
        """Creates an `EndpointConfig` instance."""
        self.url = url
        self.params = params or {}
        self.headers = headers or {}
        self.basic_auth = basic_auth or {}
        self.token = token
        self.token_name = token_name
        self.type = kwargs.pop("store_type", kwargs.pop("type", None))
        self.cafile = cafile
        self.actions_module = actions_module
        self.kwargs = kwargs

    def session(self) -> aiohttp.ClientSession:
        """Creates and returns a configured aiohttp client session."""
        # create authentication parameters
        if self.basic_auth:
            auth = aiohttp.BasicAuth(
                self.basic_auth["username"], self.basic_auth["password"]
            )
        else:
            auth = None

        return aiohttp.ClientSession(
            headers=self.headers,
            auth=auth,
            timeout=aiohttp.ClientTimeout(total=DEFAULT_REQUEST_TIMEOUT),
        )

    def combine_parameters(
        self, kwargs: Optional[Dict[Text, Any]] = None
    ) -> Dict[Text, Any]:
        # construct GET parameters
        params = self.params.copy()

        # set the authentication token if present
        if self.token:
            params[self.token_name] = self.token

        if kwargs and "params" in kwargs:
            params.update(kwargs["params"])
            del kwargs["params"]
        return params

    async def request(
        self,
        method: Text = "post",
        subpath: Optional[Text] = None,
        content_type: Optional[Text] = "application/json",
        compress: bool = False,
        **kwargs: Any,
    ) -> Optional[Any]:
        """Send a HTTP request to the endpoint. Return json response, if available.

        All additional arguments will get passed through
        to aiohttp's `session.request`.
        """
        # create the appropriate headers
        headers = {}
        if content_type:
            headers["Content-Type"] = content_type

        if "headers" in kwargs:
            headers.update(kwargs["headers"])
            del kwargs["headers"]

        if self.headers:
            headers.update(self.headers)

        url = concat_url(self.url, subpath)

        sslcontext = None
        if self.cafile:
            try:
                sslcontext = ssl.create_default_context(cafile=self.cafile)
            except FileNotFoundError as e:
                raise FileNotFoundException(
                    f"Failed to find certificate file, "
                    f"'{os.path.abspath(self.cafile)}' does not exist."
                ) from e

        async with self.session() as session:
            async with session.request(
                method,
                url,
                headers=headers,
                params=self.combine_parameters(kwargs),
                compress=compress,
                ssl=sslcontext,
                **kwargs,
            ) as response:
                if response.status >= 400:
                    raise ClientResponseError(
                        response.status,
                        response.reason,
                        await response.content.read(),
                    )
                try:
                    return await response.json()
                except ContentTypeError:
                    return None

    @classmethod
    def from_dict(cls, data: Dict[Text, Any]) -> "EndpointConfig":
        return EndpointConfig(**data)

    def copy(self) -> "EndpointConfig":
        return EndpointConfig(
            self.url,
            self.params,
            self.headers,
            self.basic_auth,
            self.token,
            self.token_name,
            **self.kwargs,
        )

    def __eq__(self, other: Any) -> bool:
        if isinstance(self, type(other)):
            return (
                other.url == self.url
                and other.params == self.params
                and other.headers == self.headers
                and other.basic_auth == self.basic_auth
                and other.token == self.token
                and other.token_name == self.token_name
            )
        else:
            return False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)


class ClientResponseError(aiohttp.ClientError):
    def __init__(self, status: int, message: Text, text: Text) -> None:
        self.status = status
        self.message = message
        self.text = text
        super().__init__(f"{status}, {message}, body='{text}'")


def bool_arg(request: Request, name: Text, default: bool = True) -> bool:
    """Returns a passed boolean argument of the request or a default.

    Checks the `name` parameter of the request if it contains a valid
    boolean value. If not, `default` is returned.

    Args:
        request: Sanic request.
        name: Name of argument.
        default: Default value for `name` argument.

    Returns:
        A bool value if `name` is a valid boolean, `default` otherwise.
    """
    return str(request.args.get(name, default)).lower() == "true"


def float_arg(
    request: Request, key: Text, default: Optional[float] = None
) -> Optional[float]:
    """Returns a passed argument cast as a float or None.

    Checks the `key` parameter of the request if it contains a valid
    float value. If not, `default` is returned.

    Args:
        request: Sanic request.
        key: Name of argument.
        default: Default value for `key` argument.

    Returns:
        A float value if `key` is a valid float, `default` otherwise.
    """
    arg = request.args.get(key, default)

    if arg is default:
        return arg

    try:
        return float(str(arg))
    except (ValueError, TypeError):
        structlogger.warning("endpoint.float_arg.convert_failed", arg=arg, key=key)
        return default


def int_arg(
    request: Request, key: Text, default: Optional[int] = None
) -> Optional[int]:
    """Returns a passed argument cast as an int or None.

    Checks the `key` parameter of the request if it contains a valid
    int value. If not, `default` is returned.

    Args:
        request: Sanic request.
        key: Name of argument.
        default: Default value for `key` argument.

    Returns:
        An int value if `key` is a valid integer, `default` otherwise.
    """
    arg = request.args.get(key, default)

    if arg is default:
        return arg

    try:
        return int(str(arg))
    except (ValueError, TypeError):
        structlogger.warning("endpoint.int_arg.convert_failed", arg=arg, key=key)
        return default



import argparse
from pathlib import Path
from typing import List, Optional, Text

import rasa.cli.arguments.evaluate as arguments
import rasa.shared.utils.cli
from rasa import telemetry
from rasa.cli import SubParsersAction
from rasa.core.evaluation.marker_base import Marker, OperatorMarker
from rasa.core.evaluation.marker_tracker_loader import MarkerTrackerLoader
from rasa.core.tracker_store import TrackerStore
from rasa.core.utils import AvailableEndpoints
from rasa.shared.core.domain import Domain

STATS_OVERALL_SUFFIX = "-overall.csv"
STATS_SESSION_SUFFIX = "-per-session.csv"


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    """Add all evaluate parsers.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
    """
    evaluate_parser = subparsers.add_parser(
        "evaluate",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Tools for evaluating models.",
    )

    evaluate_subparsers = evaluate_parser.add_subparsers()

    marker_parser = evaluate_subparsers.add_parser(
        "markers",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Applies marker conditions to existing trackers.",
    )

    markers_subparser = marker_parser.add_subparsers(dest="strategy")

    markers_first_n_subparser = markers_subparser.add_parser(
        "first_n",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Select trackers sequentially until N are taken.",
    )
    arguments.set_markers_first_n_arguments(markers_first_n_subparser)

    arguments.set_markers_arguments(markers_first_n_subparser)

    markers_sample_subparser = markers_subparser.add_parser(
        "sample_n",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Select trackers by sampling N.",
    )
    arguments.set_markers_sample_arguments(markers_sample_subparser)

    arguments.set_markers_arguments(markers_sample_subparser)

    markers_all_subparser = markers_subparser.add_parser(
        "all",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Select all trackers.",
    )

    arguments.set_markers_arguments(markers_all_subparser)

    marker_parser.set_defaults(func=_run_markers_cli)


def _run_markers_cli(args: argparse.Namespace) -> None:
    """Run markers algorithm using parameters from CLI.

    Args:
        args: The arguments passed in from the CLI.
    """
    seed = args.seed if "seed" in args else None
    count = args.count if "count" in args else None

    stats_file_prefix = args.stats_file_prefix if args.stats else None

    _run_markers(
        seed,
        count,
        args.endpoints,
        args.domain,
        args.strategy,
        args.config,
        args.output_filename,
        stats_file_prefix,
    )


def _run_markers(
    seed: Optional[int],
    count: Optional[int],
    endpoint_config: Path,
    domain_path: Optional[Text],
    strategy: Text,
    config: Path,
    output_filename: Path,
    stats_file_prefix: Optional[Path] = None,
) -> None:
    """Run markers algorithm over specified config and tracker store.

    Args:
        seed: (Optional) The seed to initialise the random number generator for
              use with the 'sample' strategy.
        count: (Optional) Number of trackers to extract from (for any strategy
               except 'all').
        endpoint_config: Path to the endpoint configuration defining the tracker
                         store to use.
        domain_path: Path to the domain specification to use when validating the
                     marker definitions.
        strategy: Strategy to use when selecting trackers to extract from.
        config: Path to the markers definition file to use.
        output_filename: Path to write out the extracted markers.
        stats_file_prefix: (Optional) A prefix used to create paths where files with
            statistics on the marker extraction results will be written.
            It must consists of the path to the where those files should be stored
            and the common file prefix, e.g. '<path-to-stats-folder>/statistics'.
            Statistics derived from all marker extractions will be stored in
            '<path-to-stats-folder>/statistics-overall.csv', while the statistics
            computed per session will be stored in
            '<path-to-stats-folder>/statistics-per-session.csv'.
    """
    telemetry.track_markers_extraction_initiated(
        strategy=strategy,
        only_extract=stats_file_prefix is not None,
        seed=seed is not None,
        count=count,
    )

    domain = Domain.load(domain_path) if domain_path else None
    markers = Marker.from_path(config)
    if domain and not markers.validate_against_domain(domain):
        rasa.shared.utils.cli.print_error_and_exit(
            "Validation errors were found in the markers definition. "
            "Please see errors listed above and fix before running again."
        )

    # Calculate telemetry
    # All loaded markers are combined with one virtual OR over all markers
    num_markers = len(markers.sub_markers)
    max_depth = markers.max_depth() - 1
    # Find maximum branching of marker
    branching_factor = max(
        (
            len(sub_marker.sub_markers)
            for marker in markers.sub_markers
            for sub_marker in marker.flatten()
            if isinstance(sub_marker, OperatorMarker)
        ),
        default=0,
    )

    telemetry.track_markers_parsed_count(num_markers, max_depth, branching_factor)

    tracker_loader = _create_tracker_loader(
        endpoint_config, strategy, domain, count, seed
    )

    def _append_suffix(path: Optional[Path], suffix: Text) -> Optional[Path]:
        return path.parent / (path.name + suffix) if path else None

    try:
        import asyncio

        asyncio.run(
            markers.evaluate_trackers(
                trackers=tracker_loader.load(),
                output_file=output_filename,
                session_stats_file=_append_suffix(
                    stats_file_prefix, STATS_SESSION_SUFFIX
                ),
                overall_stats_file=_append_suffix(
                    stats_file_prefix, STATS_OVERALL_SUFFIX
                ),
            )
        )
    except (FileExistsError, NotADirectoryError) as e:
        rasa.shared.utils.cli.print_error_and_exit(message=str(e))


def _create_tracker_loader(
    endpoint_config: Text,
    strategy: Text,
    domain: Domain,
    count: Optional[int],
    seed: Optional[int],
) -> MarkerTrackerLoader:
    """Create a tracker loader against the configured tracker store.

    Args:
        endpoint_config: Path to the endpoint configuration defining the tracker
                         store to use.
        strategy: Strategy to use when selecting trackers to extract from.
        domain: The domain to use when connecting to the tracker store.
        count: (Optional) Number of trackers to extract from (for any strategy
               except 'all').
        seed: (Optional) The seed to initialise the random number generator for
              use with the 'sample_n' strategy.

    Returns:
        A MarkerTrackerLoader object configured with the specified strategy against
        the configured tracker store.
    """
    endpoints = AvailableEndpoints.get_instance(endpoint_config)
    tracker_store = TrackerStore.create(endpoints.tracker_store, domain=domain)
    return MarkerTrackerLoader(tracker_store, strategy, count, seed)



import asyncio
import os
from dataclasses import dataclass
from typing import Any, AsyncIterator, Dict, Optional

from rasa.core.channels.voice_stream.asr.asr_engine import ASREngine, ASREngineConfig
from rasa.core.channels.voice_stream.asr.asr_event import (
    ASREvent,
    NewTranscript,
    UserIsSpeaking,
)
from rasa.core.channels.voice_stream.audio_bytes import HERTZ, RasaAudioBytes
from rasa.shared.constants import AZURE_SPEECH_API_KEY_ENV_VAR
from rasa.shared.exceptions import ConnectionException


@dataclass
class AzureASRConfig(ASREngineConfig):
    language: Optional[str] = None
    speech_region: Optional[str] = None
    speech_host: Optional[str] = None
    speech_endpoint: Optional[str] = None


class AzureASR(ASREngine[AzureASRConfig]):
    required_env_vars = (AZURE_SPEECH_API_KEY_ENV_VAR,)
    required_packages = ("azure.cognitiveservices.speech",)

    def __init__(self, config: Optional[AzureASRConfig] = None):
        super().__init__(config)

        import azure.cognitiveservices.speech as speechsdk

        self.speech_recognizer: Optional[speechsdk.SpeechRecognizer] = None
        self.stream: Optional[speechsdk.audio.PushAudioInputStream] = None
        self.is_recognizing = False
        self.queue: asyncio.Queue[speechsdk.SpeechRecognitionEventArgs] = (
            asyncio.Queue()
        )
        self.main_loop = asyncio.get_running_loop()

    def signal_user_is_speaking(self, event: Any) -> None:
        """Replace the azure event with a generic is speaking event."""
        self.fill_queue(UserIsSpeaking())

    def fill_queue(self, event: Any) -> None:
        """Either puts the event or a dedicated ASR Event into the queue."""
        # This function is used by call backs of the azure speech library
        # which seems to run separate threads/processes
        # To properly wake up the task waiting at queue.get, we need to
        # put to the queue in the same event loop
        self.main_loop.call_soon_threadsafe(self.queue.put_nowait, event)

    async def connect(self) -> None:
        import azure.cognitiveservices.speech as speechsdk

        # connecting to eastus by default
        if (
            self.config.speech_region is None
            and self.config.speech_host is None
            and self.config.speech_endpoint is None
        ):
            self.config.speech_region = "eastus"
        speech_config = speechsdk.SpeechConfig(
            subscription=os.environ[AZURE_SPEECH_API_KEY_ENV_VAR],
            region=self.config.speech_region,
            endpoint=self.config.speech_endpoint,
            host=self.config.speech_host,
        )
        audio_format = speechsdk.audio.AudioStreamFormat(
            samples_per_second=HERTZ,
            bits_per_sample=8,
            channels=1,
            wave_stream_format=speechsdk.AudioStreamWaveFormat.MULAW,
        )
        self.stream = speechsdk.audio.PushAudioInputStream(stream_format=audio_format)
        audio_config = speechsdk.audio.AudioConfig(stream=self.stream)
        self.speech_recognizer = speechsdk.SpeechRecognizer(
            speech_config=speech_config,
            language=self.config.language,
            audio_config=audio_config,
        )
        self.speech_recognizer.recognized.connect(self.fill_queue)
        self.speech_recognizer.recognizing.connect(self.signal_user_is_speaking)
        self.speech_recognizer.start_continuous_recognition_async()
        self.is_recognizing = True

    async def close_connection(self) -> None:
        if self.speech_recognizer is None:
            raise ConnectionException("Websocket not connected.")
        self.speech_recognizer.stop_continuous_recognition_async()

    async def signal_audio_done(self) -> None:
        """Signal to the ASR Api that you are done sending data."""
        self.is_recognizing = False

    def rasa_audio_bytes_to_engine_bytes(self, chunk: RasaAudioBytes) -> bytes:
        """Convert RasaAudioBytes to bytes usable by this engine."""
        return chunk

    async def send_audio_chunks(self, chunk: RasaAudioBytes) -> None:
        """Send audio chunks to the ASR system via the websocket."""
        if self.speech_recognizer is None or self.stream is None:
            raise ConnectionException("ASR not connected.")
        engine_bytes = self.rasa_audio_bytes_to_engine_bytes(chunk)
        self.stream.write(engine_bytes)

    async def stream_asr_events(self) -> AsyncIterator[ASREvent]:
        """Stream the events returned by the ASR system as it is fed audio bytes."""
        if self.speech_recognizer is None:
            raise ConnectionException("Websocket not connected.")
        while self.is_recognizing or not self.queue.empty():
            try:
                message = await asyncio.wait_for(self.queue.get(), timeout=2)
                asr_event = self.engine_event_to_asr_event(message)
                if asr_event:
                    yield asr_event
            except asyncio.TimeoutError:
                pass

    def engine_event_to_asr_event(self, e: Any) -> Optional[ASREvent]:
        """Translate an engine event to a common ASREvent."""
        import azure.cognitiveservices.speech as speechsdk

        if isinstance(e, speechsdk.SpeechRecognitionEventArgs) and isinstance(
            e.result, speechsdk.SpeechRecognitionResult
        ):
            return NewTranscript(e.result.text)
        if isinstance(e, ASREvent):
            # transformation happened before
            return e

        return None

    @staticmethod
    def get_default_config() -> AzureASRConfig:
        return AzureASRConfig(
            language=None, speech_region=None, speech_host=None, speech_endpoint=None
        )

    @classmethod
    def from_config_dict(cls, config: Dict) -> "AzureASR":
        return AzureASR(AzureASRConfig.from_dict(config))



import argparse
import webbrowser
from asyncio import AbstractEventLoop
from typing import List, Optional, Text

from sanic import Sanic

from rasa import telemetry
from rasa.cli import SubParsersAction
from rasa.cli.arguments import shell as arguments
from rasa.core import constants
from rasa.engine.storage.local_model_storage import LocalModelStorage
from rasa.exceptions import ModelNotFound
from rasa.model import get_local_model
from rasa.shared.utils.cli import print_error
from rasa.utils.cli import remove_argument_from_parser


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    """Add all inspect parsers.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
    """
    inspect_parser = subparsers.add_parser(
        "inspect",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help=(
            "Loads your trained model and lets you talk to your "
            "assistant in the browser."
        ),
    )
    inspect_parser.set_defaults(func=inspect)
    arguments.set_shell_arguments(inspect_parser)

    # additional argument for voice
    inspect_parser.add_argument(
        "--voice", help="Enable voice", action="store_true", default=False
    )

    # it'd be confusing to expose those arguments to the user,
    # so we remove them
    remove_argument_from_parser(inspect_parser, "--credentials")
    remove_argument_from_parser(inspect_parser, "--connector")
    remove_argument_from_parser(inspect_parser, "--enable-api")


async def open_inspector_in_browser(
    server_url: Text,
    voice: bool = False,
    token: Optional[Text] = None,
) -> None:
    """Opens the rasa inspector in the default browser."""
    channel = "socketio" if not voice else "browser_audio"
    webbrowser.open(f"{server_url}/webhooks/{channel}/inspect.html?token={token}")


def inspect(args: argparse.Namespace) -> None:
    """Inspect the bot using the most recent model."""
    import rasa.cli.run
    from rasa.cli.utils import get_validated_path
    from rasa.shared.constants import DEFAULT_MODELS_PATH

    async def after_start_hook_open_inspector(_: Sanic, __: AbstractEventLoop) -> None:
        """Hook to open the browser on server start."""
        server_url = constants.DEFAULT_SERVER_FORMAT.format("http", args.port)
        await open_inspector_in_browser(server_url, args.voice, args.auth_token)

    # the following arguments are not exposed to the user
    if args.voice:
        args.connector = "browser_audio"
    else:
        args.connector = "rasa.core.channels.socketio.SocketIOInput"
    args.enable_api = True
    args.inspect = True
    args.credentials = None
    args.server_listeners = [(after_start_hook_open_inspector, "after_server_start")]

    model = get_validated_path(args.model, "model", DEFAULT_MODELS_PATH)

    try:
        model = get_local_model(model)
    except ModelNotFound:
        print_error(
            "No model found. Train a model before running the "
            "server using `rasa train`."
        )
        return

    metadata = LocalModelStorage.metadata_from_archive(model)

    telemetry.track_inspect_started(args.connector, metadata.assistant_id)
    rasa.cli.run.run(args)



import datetime
import json
import logging
import re
from http import HTTPStatus
from typing import Any, Awaitable, Callable, Dict, Iterable, List, Optional, Text

import jwt
import requests
from jwt import InvalidKeyError, PyJWTError
from jwt.algorithms import RSAAlgorithm
from requests import HTTPError
from sanic import Blueprint, response
from sanic.request import Request
from sanic.response import HTTPResponse

from rasa.core.channels.channel import InputChannel, OutputChannel, UserMessage

MICROSOFT_OPEN_ID_URI = (
    "https://login.botframework.com/v1/.well-known/openidconfiguration"
)

BEARER_REGEX = re.compile(r"Bearer\s+(.*)")

logger = logging.getLogger(__name__)

MICROSOFT_OAUTH2_URL = "https://login.microsoftonline.com"

MICROSOFT_OAUTH2_PATH = "botframework.com/oauth2/v2.0/token"


class BotFramework(OutputChannel):
    """A Microsoft Bot Framework communication channel."""

    token_expiration_date = datetime.datetime.now()

    headers = None

    @classmethod
    def name(cls) -> Text:
        return "botframework"

    def __init__(
        self,
        app_id: Text,
        app_password: Text,
        conversation: Dict[Text, Any],
        bot: Text,
        service_url: Text,
    ) -> None:
        service_url = (
            f"{service_url}/" if not service_url.endswith("/") else service_url
        )

        self.app_id = app_id
        self.app_password = app_password
        self.conversation = conversation
        self.global_uri = f"{service_url}v3/"
        self.bot = bot

    async def _get_headers(self) -> Optional[Dict[Text, Any]]:
        if BotFramework.token_expiration_date < datetime.datetime.now():
            uri = f"{MICROSOFT_OAUTH2_URL}/{MICROSOFT_OAUTH2_PATH}"
            grant_type = "client_credentials"
            scope = "https://api.botframework.com/.default"
            payload = {
                "client_id": self.app_id,
                "client_secret": self.app_password,
                "grant_type": grant_type,
                "scope": scope,
            }

            token_response = requests.post(uri, data=payload)

            if token_response.ok:
                token_data = token_response.json()
                access_token = token_data["access_token"]
                token_expiration = token_data["expires_in"]

                delta = datetime.timedelta(seconds=int(token_expiration))
                BotFramework.token_expiration_date = datetime.datetime.now() + delta

                BotFramework.headers = {
                    "content-type": "application/json",
                    "Authorization": "Bearer %s" % access_token,
                }
                return BotFramework.headers
            else:
                logger.error("Could not get BotFramework token")
                return None
        else:
            return BotFramework.headers

    def prepare_message(
        self, recipient_id: Text, message_data: Dict[Text, Any]
    ) -> Dict[Text, Any]:
        data = {
            "type": "message",
            "recipient": {"id": recipient_id},
            "from": self.bot,
            "channelData": {"notification": {"alert": "true"}},
            "text": "",
        }
        data.update(message_data)
        return data

    async def send(self, message_data: Dict[Text, Any]) -> None:
        post_message_uri = "{}conversations/{}/activities".format(
            self.global_uri, self.conversation["id"]
        )
        headers = await self._get_headers()
        send_response = requests.post(
            post_message_uri, headers=headers, data=json.dumps(message_data)
        )

        if not send_response.ok:
            logger.error(
                "Error trying to send botframework messge. Response: %s",
                send_response.text,
            )

    async def send_text_message(
        self, recipient_id: Text, text: Text, **kwargs: Any
    ) -> None:
        for message_part in text.strip().split("\n\n"):
            text_message = {"text": message_part}
            message = self.prepare_message(recipient_id, text_message)
            await self.send(message)

    async def send_image_url(
        self, recipient_id: Text, image: Text, **kwargs: Any
    ) -> None:
        hero_content = {
            "contentType": "application/vnd.microsoft.card.hero",
            "content": {"images": [{"url": image}]},
        }

        image_message = {"attachments": [hero_content]}
        message = self.prepare_message(recipient_id, image_message)
        await self.send(message)

    async def send_text_with_buttons(
        self,
        recipient_id: Text,
        text: Text,
        buttons: List[Dict[Text, Any]],
        **kwargs: Any,
    ) -> None:
        hero_content = {
            "contentType": "application/vnd.microsoft.card.hero",
            "content": {"subtitle": text, "buttons": buttons},
        }

        buttons_message = {"attachments": [hero_content]}
        message = self.prepare_message(recipient_id, buttons_message)
        await self.send(message)

    async def send_elements(
        self, recipient_id: Text, elements: Iterable[Dict[Text, Any]], **kwargs: Any
    ) -> None:
        for e in elements:
            message = self.prepare_message(recipient_id, e)
            await self.send(message)

    async def send_custom_json(
        self, recipient_id: Text, json_message: Dict[Text, Any], **kwargs: Any
    ) -> None:
        json_message.setdefault("type", "message")
        json_message.setdefault("recipient", {}).setdefault("id", recipient_id)
        json_message.setdefault("from", self.bot)
        json_message.setdefault("channelData", {}).setdefault(
            "notification", {}
        ).setdefault("alert", "true")
        json_message.setdefault("text", "")
        await self.send(json_message)


class BotFrameworkInput(InputChannel):
    """Bot Framework input channel implementation."""

    @classmethod
    def name(cls) -> Text:
        return "botframework"

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> InputChannel:
        if not credentials:
            cls.raise_missing_credentials_exception()

        return cls(credentials.get("app_id"), credentials.get("app_password"))

    def __init__(self, app_id: Text, app_password: Text) -> None:
        """Create a Bot Framework input channel.

        Args:
            app_id: Bot Framework's API id
            app_password: Bot Framework application secret
        """
        self.app_id = app_id
        self.app_password = app_password

        self.jwt_keys: Dict[Text, Any] = {}
        self.jwt_update_time = datetime.datetime.fromtimestamp(0)

        self._update_cached_jwk_keys()

    def _update_cached_jwk_keys(self) -> None:
        logger.debug("Updating JWT keys for the Botframework.")
        response = requests.get(MICROSOFT_OPEN_ID_URI)
        response.raise_for_status()
        conf = response.json()

        jwks_uri = conf["jwks_uri"]

        keys_request = requests.get(jwks_uri)
        keys_request.raise_for_status()
        keys_list = keys_request.json()
        self.jwt_keys = {key["kid"]: key for key in keys_list["keys"]}
        self.jwt_update_time = datetime.datetime.now()

    def _validate_jwt_token(self, jwt_token: Text) -> None:
        jwt_header = jwt.get_unverified_header(jwt_token)
        key_id = jwt_header["kid"]
        if key_id not in self.jwt_keys:
            raise InvalidKeyError(f"JWT Key with ID {key_id} not found.")

        key_json = self.jwt_keys[key_id]
        public_key = RSAAlgorithm.from_jwk(key_json)
        jwt.decode(
            jwt_token,
            key=public_key,
            audience=self.app_id,
            algorithms=jwt_header["alg"],
        )

    def _validate_auth(self, auth_header: Optional[Text]) -> Optional[HTTPResponse]:
        if not auth_header:
            return response.text(
                "No authorization header provided.", status=HTTPStatus.UNAUTHORIZED
            )

        # Update the JWT keys daily
        if datetime.datetime.now() - self.jwt_update_time > datetime.timedelta(days=1):
            try:
                self._update_cached_jwk_keys()
            except HTTPError as error:
                logger.warning(
                    f"Could not update JWT keys from {MICROSOFT_OPEN_ID_URI}."
                )
                logger.exception(error, exc_info=True)

        auth_match = BEARER_REGEX.match(auth_header)
        if not auth_match:
            return response.text(
                "No Bearer token provided in Authorization header.",
                status=HTTPStatus.UNAUTHORIZED,
            )

        (jwt_token,) = auth_match.groups()

        try:
            self._validate_jwt_token(jwt_token)
        except PyJWTError as error:
            logger.error("Bot framework JWT token could not be verified.")
            logger.exception(error, exc_info=True)
            return response.text(
                "Could not validate JWT token.", status=HTTPStatus.UNAUTHORIZED
            )

        return None

    @staticmethod
    def add_attachments_to_metadata(
        postdata: Dict[Text, Any], metadata: Optional[Dict[Text, Any]]
    ) -> Optional[Dict[Text, Any]]:
        """Merge the values of `postdata['attachments']` with `metadata`."""
        if postdata.get("attachments"):
            attachments = {"attachments": postdata["attachments"]}
            if metadata:
                metadata.update(attachments)
            else:
                metadata = attachments
        return metadata

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> Blueprint:
        """Defines the Sanic blueprint for the bot framework integration."""
        botframework_webhook = Blueprint("botframework_webhook", __name__)

        @botframework_webhook.route("/", methods=["GET"])
        async def health(request: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @botframework_webhook.route("/webhook", methods=["POST"])
        async def webhook(request: Request) -> HTTPResponse:
            validation_response = self._validate_auth(
                request.headers.get("Authorization")
            )
            if validation_response:
                return validation_response

            postdata = request.json
            metadata = self.get_metadata(request)

            metadata_with_attachments = self.add_attachments_to_metadata(
                postdata, metadata
            )

            try:
                if postdata["type"] == "message":
                    out_channel = BotFramework(
                        self.app_id,
                        self.app_password,
                        postdata["conversation"],
                        postdata["recipient"],
                        postdata["serviceUrl"],
                    )

                    user_msg = UserMessage(
                        text=postdata.get("text", ""),
                        output_channel=out_channel,
                        sender_id=postdata["from"]["id"],
                        input_channel=self.name(),
                        metadata=metadata_with_attachments,
                    )

                    await on_new_message(user_msg)
                else:
                    logger.info("Not received message type")
            except Exception as e:
                logger.error(f"Exception when trying to handle message.{e}")
                logger.debug(e, exc_info=True)
                pass

            return response.text("success")

        return botframework_webhook



from __future__ import annotations

from typing import Any, Dict, Text

from rasa.shared.core.flows.flow_step import FlowStep


class InternalFlowStep(FlowStep):
    """A superclass for built-in flow steps.

    Built-in flow steps are required to manage the lifecycle of a
    flow and are not intended to be used by users.
    """

    @classmethod
    def from_json(cls, flow_id: Text, data: Dict[Text, Any]) -> InternalFlowStep:
        """Create an InternalFlowStep object from serialized data.

        Args:
            flow_id: The id of the flow that contains the step.
            data: data for an InternalFlowStep in a serialized format

        Returns:
            Raises because InternalFlowSteps are not serialized or de-serialized.
        """
        raise ValueError(
            "Internal flow steps are ephemeral and are not to be serialized "
            "or de-serialized."
        )

    def as_json(self) -> Dict[Text, Any]:
        """Serialize the InternalFlowStep object

        Returns:
            Raises because InternalFlowSteps are not serialized or de-serialized.
        """
        raise ValueError(
            "Internal flow steps are ephemeral and are not to be serialized "
            "or de-serialized."
        )

    @property
    def default_id_postfix(self) -> str:
        """Returns the default id postfix of the flow step."""
        raise ValueError("Internal flow steps do not need a default id")



import functools
import json
import logging
from typing import TYPE_CHECKING, Callable, List, Optional, Type

from opentelemetry.trace import Tracer

if TYPE_CHECKING:
    from rasa.core.policies.intentless_policy import IntentlessPolicy

logger = logging.getLogger(__name__)


def _instrument_select_response_examples(
    tracer: Tracer, policy_class: Type["IntentlessPolicy"]
) -> None:
    def tracing_select_response_examples_wrapper(fn: Callable) -> Callable:
        @functools.wraps(fn)
        def wrapper(
            self: "IntentlessPolicy",
            history: str,
            number_of_samples: int,
            max_number_of_tokens: int,
        ) -> List[str]:
            with tracer.start_as_current_span(
                f"{self.__class__.__name__}.{fn.__name__}"
            ) as span:
                ai_response_examples = fn(
                    self, history, number_of_samples, max_number_of_tokens
                )
                span.set_attributes(
                    {
                        "ai_response_examples": json.dumps(ai_response_examples),
                    }
                )
                return ai_response_examples

        return wrapper

    policy_class.select_response_examples = tracing_select_response_examples_wrapper(  # type: ignore[assignment]
        policy_class.select_response_examples
    )

    logger.debug(
        f"Instrumented '{policy_class.__name__}.select_response_examples' method."
    )


def _instrument_select_few_shot_conversations(
    tracer: Tracer, policy_class: Type["IntentlessPolicy"]
) -> None:
    def tracing_select_few_shot_conversations_wrapper(fn: Callable) -> Callable:
        @functools.wraps(fn)
        def wrapper(
            self: "IntentlessPolicy",
            history: str,
            number_of_samples: int,
            max_number_of_tokens: int,
        ) -> List[str]:
            with tracer.start_as_current_span(
                f"{self.__class__.__name__}.{fn.__name__}"
            ) as span:
                conversation_samples = fn(
                    self, history, number_of_samples, max_number_of_tokens
                )
                span.set_attributes(
                    {
                        "conversation_samples": json.dumps(conversation_samples),
                    }
                )
                return conversation_samples

        return wrapper

    policy_class.select_few_shot_conversations = (  # type: ignore[assignment]
        tracing_select_few_shot_conversations_wrapper(
            policy_class.select_few_shot_conversations
        )
    )

    logger.debug(
        f"Instrumented '{policy_class.__name__}.select_few_shot_conversations' method."
    )


def _instrument_extract_ai_responses(
    tracer: Tracer, policy_class: Type["IntentlessPolicy"]
) -> None:
    def tracing_extract_ai_responses_wrapper(fn: Callable) -> Callable:
        @functools.wraps(fn)
        def wrapper(
            self: "IntentlessPolicy", conversation_samples: List[str]
        ) -> List[str]:
            with tracer.start_as_current_span(
                f"{self.__class__.__name__}.{fn.__name__}"
            ) as span:
                ai_responses = fn(self, conversation_samples)
                span.set_attributes(
                    {
                        "ai_responses": json.dumps(ai_responses),
                    }
                )
                return ai_responses

        return wrapper

    policy_class.extract_ai_responses = tracing_extract_ai_responses_wrapper(  # type: ignore[assignment]
        policy_class.extract_ai_responses
    )

    logger.debug(f"Instrumented '{policy_class.__name__}.extract_ai_responses' method.")


def _instrument_generate_answer(
    tracer: Tracer, policy_class: Type["IntentlessPolicy"]
) -> None:
    def tracing_generate_answer_wrapper(fn: Callable) -> Callable:
        @functools.wraps(fn)
        async def wrapper(
            self: "IntentlessPolicy",
            response_examples: List[str],
            conversation_samples: List[str],
            history: str,
        ) -> Optional[str]:
            with tracer.start_as_current_span(
                f"{self.__class__.__name__}.{fn.__name__}"
            ) as span:
                llm_response = await fn(
                    self, response_examples, conversation_samples, history
                )
                span.set_attributes(
                    {
                        "llm_response": json.dumps(llm_response),
                    }
                )
                return llm_response

        return wrapper

    policy_class.generate_answer = tracing_generate_answer_wrapper(  # type: ignore[assignment]
        policy_class.generate_answer
    )

    logger.debug(f"Instrumented '{policy_class.__name__}.generate_answer' method.")



from __future__ import annotations

import base64
import json
import uuid
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, Optional, Text, Tuple

import structlog
from sanic import (  # type: ignore[attr-defined]
    Blueprint,
    HTTPResponse,
    Request,
    Websocket,
    response,
)

from rasa.core.channels import InputChannel, UserMessage
from rasa.core.channels.channel import (
    create_auth_requested_response_provider,
    requires_basic_auth,
)
from rasa.core.channels.voice_ready.utils import CallParameters
from rasa.core.channels.voice_stream.audio_bytes import RasaAudioBytes
from rasa.core.channels.voice_stream.call_state import call_state
from rasa.core.channels.voice_stream.tts.tts_engine import TTSEngine
from rasa.core.channels.voice_stream.voice_channel import (
    ContinueConversationAction,
    EndConversationAction,
    NewAudioAction,
    VoiceChannelAction,
    VoiceInputChannel,
    VoiceOutputChannel,
)
from rasa.shared.exceptions import RasaException

if TYPE_CHECKING:
    from twilio.twiml.voice_response import VoiceResponse

logger = structlog.get_logger(__name__)


TWILIO_MEDIA_STREAMS_WEBHOOK_PATH = "webhooks/twilio_media_streams/webhook"
TWILIO_MEDIA_STREAMS_WEBSOCKET_PATH = "webhooks/twilio_media_streams/websocket"


CALL_SID_REQUEST_KEY = "CallSid"
FROM_NUMBER_REQUEST_KEY = "From"
TO_NUMBER_REQUEST_KEY = "To"
DIRECTION_REQUEST_KEY = "Direction"


def map_call_params(data: Dict[Text, Any]) -> CallParameters:
    """Map the twilio stream parameters to the CallParameters dataclass."""
    stream_sid = data["streamSid"]
    parameters = data["start"]["customParameters"]
    return CallParameters(
        call_id=parameters.get("call_id", ""),
        user_phone=parameters.get("user_phone", ""),
        bot_phone=parameters.get("bot_phone", ""),
        direction=parameters.get("direction"),
        stream_id=stream_sid,
    )


class TwilioMediaStreamsOutputChannel(VoiceOutputChannel):
    @classmethod
    def name(cls) -> str:
        return "twilio_media_streams"

    def rasa_audio_bytes_to_channel_bytes(
        self, rasa_audio_bytes: RasaAudioBytes
    ) -> bytes:
        return base64.b64encode(rasa_audio_bytes)

    def create_marker_message(self, recipient_id: str) -> Tuple[str, str]:
        message_id = uuid.uuid4().hex
        mark_message = json.dumps(
            {
                "event": "mark",
                "streamSid": recipient_id,
                "mark": {"name": message_id},
            }
        )
        return mark_message, message_id

    def channel_bytes_to_message(self, recipient_id: str, channel_bytes: bytes) -> str:
        media_message = json.dumps(
            {
                "event": "media",
                "streamSid": recipient_id,
                "media": {
                    "payload": channel_bytes.decode("utf-8"),
                },
            }
        )
        return media_message


class TwilioMediaStreamsInputChannel(VoiceInputChannel):
    def __init__(
        self,
        server_url: str,
        asr_config: Dict,
        tts_config: Dict,
        monitor_silence: bool = False,
        username: Optional[Text] = None,
        password: Optional[Text] = None,
    ):
        super().__init__(server_url, asr_config, tts_config, monitor_silence)
        self.username = username
        self.password = password

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[str, Any]]) -> InputChannel:
        credentials = credentials or {}

        username = credentials.get("username")
        password = credentials.get("password")
        if (username is None) != (password is None):
            raise RasaException(
                "In TwilioMediaStreams channel, either both username and password "
                "or neither should be provided. "
            )

        return cls(
            credentials["server_url"],
            credentials["asr"],
            credentials["tts"],
            credentials.get("monitor_silence", False),
            username=username,
            password=password,
        )

    @classmethod
    def name(cls) -> str:
        return "twilio_media_streams"

    def channel_bytes_to_rasa_audio_bytes(self, input_bytes: bytes) -> RasaAudioBytes:
        return RasaAudioBytes(base64.b64decode(input_bytes))

    async def collect_call_parameters(
        self, channel_websocket: Websocket
    ) -> Optional[CallParameters]:
        async for message in channel_websocket:
            data = json.loads(message)
            if data["event"] == "start":
                # retrieve parameters set in the webhook - contains info about the
                # caller
                return map_call_params(data)
        return None

    def map_input_message(
        self,
        message: Any,
        ws: Websocket,
    ) -> VoiceChannelAction:
        data = json.loads(message)
        if data["event"] == "media":
            audio_bytes = self.channel_bytes_to_rasa_audio_bytes(
                data["media"]["payload"]
            )
            return NewAudioAction(audio_bytes)
        elif data["event"] == "stop":
            return EndConversationAction()
        elif data["event"] == "mark":
            if data["mark"]["name"] == call_state.latest_bot_audio_id:
                # Just finished streaming last audio bytes
                call_state.is_bot_speaking = False  # type: ignore[attr-defined]
                if call_state.should_hangup:
                    logger.debug(
                        "twilio_streams.hangup", marker=call_state.latest_bot_audio_id
                    )
                    return EndConversationAction()
            else:
                call_state.is_bot_speaking = True  # type: ignore[attr-defined]
        return ContinueConversationAction()

    def create_output_channel(
        self, voice_websocket: Websocket, tts_engine: TTSEngine
    ) -> VoiceOutputChannel:
        return TwilioMediaStreamsOutputChannel(
            voice_websocket,
            tts_engine,
            self.tts_cache,
        )

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> Blueprint:
        """Defines a Sanic blueprint for the voice input channel."""
        blueprint = Blueprint("twilio_media_streams", __name__)

        @blueprint.route("/", methods=["GET"])
        async def health(_: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @blueprint.route("/webhook", methods=["POST"])
        @requires_basic_auth(
            username=self.username,
            password=self.password,
            auth_request_provider=create_auth_requested_response_provider(
                realm=TWILIO_MEDIA_STREAMS_WEBHOOK_PATH
            ),
        )
        async def receive(request: Request) -> HTTPResponse:
            voice_response = self._build_twilio_response(request)

            logger.debug(
                "twilio_media_streams.webhook.twilio_response",
                twilio_response=str(voice_response),
            )

            return response.text(str(voice_response), content_type="text/xml")

        @blueprint.websocket("/websocket")  # type: ignore
        async def handle_message(request: Request, ws: Websocket) -> None:
            await self.run_audio_streaming(on_new_message, ws)

        return blueprint

    def _websocket_stream_url(self) -> str:
        """Returns the websocket stream URL."""
        # depending on the config value, the url might contain http as a
        # protocol or not - we'll make sure both work
        if self.server_url.startswith("http"):
            base_url = self.server_url.replace("http", "ws")
        else:
            base_url = f"wss://{self.server_url}"
        return f"{base_url}/{TWILIO_MEDIA_STREAMS_WEBSOCKET_PATH}"

    def _build_twilio_response(self, request: Request) -> VoiceResponse:
        from twilio.twiml.voice_response import Connect, VoiceResponse

        voice_response = VoiceResponse()
        start = Connect()
        stream = start.stream(url=self._websocket_stream_url())
        # pass information about the call to the webhook - so we can
        # store it in the input channel
        stream.parameter(
            name="call_id", value=request.form.get(CALL_SID_REQUEST_KEY, None)
        )
        stream.parameter(
            name="user_phone", value=request.form.get(FROM_NUMBER_REQUEST_KEY, None)
        )
        stream.parameter(
            name="bot_phone", value=request.form.get(TO_NUMBER_REQUEST_KEY, None)
        )
        stream.parameter(
            name="direction", value=request.form.get(DIRECTION_REQUEST_KEY, None)
        )
        voice_response.append(start)
        return voice_response



import json
from typing import TYPE_CHECKING, Any, Dict, List, Tuple
from urllib.parse import urlparse

import grpc
import structlog
from google.protobuf.json_format import MessageToDict, Parse, ParseDict
from rasa_sdk.grpc_errors import ResourceNotFound, ResourceNotFoundType
from rasa_sdk.grpc_py import action_webhook_pb2, action_webhook_pb2_grpc

from rasa.core.actions.action_exceptions import DomainNotFound
from rasa.core.actions.constants import SSL_CLIENT_CERT_FIELD, SSL_CLIENT_KEY_FIELD
from rasa.core.actions.custom_action_executor import (
    CustomActionExecutor,
    CustomActionRequestWriter,
)
from rasa.shared.exceptions import RasaException
from rasa.shared.utils.io import file_as_bytes
from rasa.utils.endpoints import EndpointConfig

if TYPE_CHECKING:
    from rasa.shared.core.domain import Domain
    from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger(__name__)


class GRPCCustomActionExecutor(CustomActionExecutor):
    """gRPC-based implementation of the CustomActionExecutor.

    Executes custom actions by making gRPC requests to the action endpoint.
    """

    def __init__(
        self,
        action_name: str,
        action_endpoint: EndpointConfig,
    ) -> None:
        """Initializes the gRPC custom action executor.

        Args:
            action_name: Name of the custom action.
            action_endpoint: Endpoint configuration of the custom action.
        """
        self.action_name = action_name
        self.request_writer = CustomActionRequestWriter(action_name, action_endpoint)
        self.action_endpoint = action_endpoint

        parsed_url = urlparse(self.action_endpoint.url)
        self.request_url = parsed_url.netloc

        self.cert_ca = (
            file_as_bytes(self.action_endpoint.cafile)
            if self.action_endpoint.cafile
            else None
        )

        self.client_cert = None
        self.client_key = None

        client_cert_file = self.action_endpoint.kwargs.get(SSL_CLIENT_CERT_FIELD)
        client_key_file = self.action_endpoint.kwargs.get(SSL_CLIENT_KEY_FIELD)
        if client_cert_file and client_key_file:
            self.client_cert = file_as_bytes(client_cert_file)
            self.client_key = file_as_bytes(client_key_file)
        elif client_key_file and not client_cert_file:
            structlogger.error(
                f"rasa.core.actions.grpc_custom_action_executor.{SSL_CLIENT_CERT_FIELD}_missing",
                event_info=(
                    f"Client key file '{SSL_CLIENT_KEY_FIELD}' is provided but "
                    f"client certificate file '{SSL_CLIENT_CERT_FIELD}' "
                    f"is not provided in the endpoint configuration. "
                    f"Both fields are required for client TLS authentication."
                    f"Continuing without client TLS authentication."
                ),
            )
        elif client_cert_file and not client_key_file:
            structlogger.error(
                f"rasa.core.actions.grpc_custom_action_executor.{SSL_CLIENT_KEY_FIELD}_missing",
                event_info=(
                    f"Client certificate file '{SSL_CLIENT_CERT_FIELD}' "
                    f" is provided but client key file '{SSL_CLIENT_KEY_FIELD}'"
                    f" is not provided in the endpoint configuration. "
                    f"Both fields are required for client TLS authentication."
                    f"Continuing without client TLS authentication."
                ),
            )

    async def run(
        self,
        tracker: "DialogueStateTracker",
        domain: "Domain",
        include_domain: bool = False,
    ) -> Dict[str, Any]:
        """Execute the custom action using a gRPC request.

        Args:
            tracker: Tracker for the current conversation.
            domain: Domain of the assistant.
            include_domain: If True, the domain is included in the request.

        Returns:
            Response from the action server.
        """

        request = self._create_payload(
            tracker=tracker, domain=domain, include_domain=include_domain
        )

        return self._request(request)

    def _request(
        self,
        request: action_webhook_pb2.WebhookRequest,
    ) -> Dict[str, Any]:
        """Perform a single gRPC request to the action server.

        Args:
            request: gRPC Request to be sent to the action server.

        Returns:
            Response from the action server.
        """

        client = self._create_grpc_client()
        metadata = self._build_metadata()
        try:
            response = client.Webhook(request, metadata=metadata)
            return MessageToDict(response)
        except grpc.RpcError as rpc_error:
            if not isinstance(rpc_error, grpc.Call):
                raise RasaException(
                    f"Failed to execute custom action '{self.action_name}'. "
                    f"Unknown error occurred while calling the "
                    f"action server over gRPC protocol."
                )

            status_code = rpc_error.code()
            details = rpc_error.details()
            if status_code is not grpc.StatusCode.NOT_FOUND:
                raise RasaException(
                    f"Failed to execute custom action '{self.action_name}'. "
                    f"Error: {details}"
                )

            resource_not_found_error = ResourceNotFound.model_validate_json(details)
            if (
                resource_not_found_error
                and resource_not_found_error.resource_type
                == ResourceNotFoundType.DOMAIN
            ):
                structlogger.error(
                    "rasa.core.actions.grpc_custom_action_executor.domain_not_found",
                    event_info=(
                        f"Failed to execute custom action '{self.action_endpoint}'. "
                        f"Could not find domain. {resource_not_found_error.message}"
                    ),
                )
                raise DomainNotFound()
            raise RasaException(
                f"Failed to execute custom action '{self.action_name}'. "
                f"Error: {details}"
            )

    def _build_metadata(self) -> List[Tuple[str, Any]]:
        """Build metadata for the gRPC request.

        Returns:
            Metadata for the gRPC request.
        """
        metadata = []
        for key, value in self.action_endpoint.headers.items():
            metadata.append((key, value))
        return metadata

    def _create_payload(
        self,
        tracker: "DialogueStateTracker",
        domain: "Domain",
        include_domain: bool = False,
    ) -> action_webhook_pb2.WebhookRequest:
        """Create the gRPC payload for the action server.

        Args:
            tracker: Tracker for the current conversation.
            domain: Domain of the assistant.
            include_domain: If True, the domain is included in the request.

        Returns:
            gRPC payload for the action server.
        """
        json_body = self.request_writer.create(
            tracker=tracker, domain=domain, include_domain=include_domain
        )

        request_proto = action_webhook_pb2.WebhookRequest()

        try:
            return ParseDict(
                js_dict=json_body, message=request_proto, ignore_unknown_fields=True
            )
        except Exception:
            structlogger.warning(
                (
                    "rasa.core.actions.grpc_custom_action_executor."
                    "create_grpc_payload_from_dict_failed"
                ),
                event_info=(
                    "Failed to create gRPC payload from Python dictionary. "
                    "Falling back to create payload from JSON intermediary."
                ),
            )
            json_text = json.dumps(json_body)
            return Parse(
                text=json_text, message=request_proto, ignore_unknown_fields=True
            )

    def _create_grpc_client(
        self,
    ) -> action_webhook_pb2_grpc.ActionServiceStub:
        """Create a gRPC client for the action server.

        Returns:
            gRPC client for the action server.
        """
        channel = self._create_channel()
        return action_webhook_pb2_grpc.ActionServiceStub(channel)

    def _create_channel(
        self,
    ) -> grpc.Channel:
        """Create a gRPC channel for the action server.

        Returns:
            gRPC channel for the action server.
        """

        compression = grpc.Compression.Gzip

        if self.cert_ca:
            credentials = grpc.ssl_channel_credentials(
                root_certificates=self.cert_ca,
                private_key=self.client_key,
                certificate_chain=self.client_cert,
            )
            return grpc.secure_channel(
                target=self.request_url,
                credentials=credentials,
                compression=compression,
            )
        return grpc.insecure_channel(target=self.request_url, compression=compression)



import logging
from typing import Iterable, Optional, Text

from rasa.core.brokers.broker import EventBroker
from rasa.core.secrets_manager.secret_manager import EndpointResolver
from rasa.core.tracker_store import TrackerStore, create_tracker_store
from rasa.shared.core.domain import Domain
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.utils.endpoints import EndpointConfig

logger = logging.getLogger(__name__)

DEFAULT_RETRIES = 3


class AuthRetryTrackerStore(TrackerStore):
    """Tracker store wrapper which implements retry mechanism in every abstract method.

    The retry mechanism first updates credentials from the secret manager.
    """

    def __init__(
        self,
        domain: "Domain",
        endpoint_config: "EndpointConfig",
        retries: int = DEFAULT_RETRIES,
        event_broker: Optional["EventBroker"] = None,
    ) -> None:
        """Initialise the tracker store wrapper."""
        self.retries = self.validate_retries(retries)
        self.endpoint_config = endpoint_config

        self._tracker_store = self.recreate_tracker_store(domain, event_broker)

        super().__init__(domain, event_broker)

    @property
    def domain(self) -> Domain:
        """Returns the domain of the wrapped tracker store."""
        return self._tracker_store.domain

    @domain.setter
    def domain(self, domain: Optional[Domain]) -> None:
        """Sets the domain of wrapped tracker store."""
        self._tracker_store.domain = domain or Domain.empty()

    @staticmethod
    def validate_retries(retries: int) -> int:
        """Validate the number of retries."""
        if retries <= 0:
            logger.warning(
                f"Invalid number of retries: {retries}. "
                f"Using default number of retries: {DEFAULT_RETRIES}."
            )
            return DEFAULT_RETRIES
        else:
            return retries

    async def keys(self) -> Iterable[Text]:
        """Retries retrieving the keys if it fails."""
        # add + 1 to retries because the retries are additional to the first attempt
        for _ in range(self.retries + 1):
            try:
                return await self._tracker_store.keys()
            except Exception as e:
                logger.warning("Failed to retrieve keys. Retrying...", exc_info=e)
                self._tracker_store = self.recreate_tracker_store(
                    self.domain, self.event_broker
                )
        else:
            logger.error(f"Failed to retrieve keys after {self.retries} retries.")
            return []

    async def retrieve(self, sender_id: Text) -> Optional["DialogueStateTracker"]:
        """Retries retrieving the tracker if it fails."""
        # add + 1 to retries because the retries are additional to the first attempt
        for _ in range(self.retries + 1):
            try:
                return await self._tracker_store.retrieve(sender_id)
            except Exception as e:
                logger.warning(
                    f"Failed to retrieve tracker for {sender_id}. Retrying...",
                    exc_info=e,
                )
                self._tracker_store = self.recreate_tracker_store(
                    self.domain, self.event_broker
                )
        else:
            logger.error(
                f"Failed to retrieve tracker for {sender_id} "
                f"after {self.retries} retries."
            )
            return None

    async def save(self, tracker: "DialogueStateTracker") -> None:
        """Retries saving the tracker if it fails."""
        # add + 1 to retries because the retries are additional to the first attempt
        for _ in range(self.retries + 1):
            try:
                await self._tracker_store.save(tracker)
                break
            except Exception as e:
                logger.warning(
                    f"Failed to save tracker for {tracker.sender_id}. Retrying...",
                    exc_info=e,
                )
                self._tracker_store = self.recreate_tracker_store(
                    self.domain, self.event_broker
                )
        else:
            logger.error(
                f"Failed to save tracker for {tracker.sender_id} "
                f"after {self.retries} retries."
            )

    def recreate_tracker_store(
        self, domain: "Domain", event_broker: Optional["EventBroker"] = None
    ) -> TrackerStore:
        """Recreate tracker store with updated credentials."""
        endpoint_config = EndpointResolver.update_config(self.endpoint_config)
        return create_tracker_store(endpoint_config, domain, event_broker)



from __future__ import annotations

import asyncio
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Text,
)

import orjson
import structlog
from sanic import (  # type: ignore[attr-defined]
    Blueprint,
    Sanic,
    Websocket,
    exceptions,
    response,
)
from sanic.request import Request
from socketio import AsyncServer

from rasa.core.channels.channel import InputChannel, OutputChannel
from rasa.shared.core.trackers import EventVerbosity
from rasa.shared.utils.cli import print_info

if TYPE_CHECKING:
    from sanic.request import Request
    from sanic.response import HTTPResponse

    from rasa.core.channels.channel import UserMessage
    from rasa.shared.core.trackers import DialogueStateTracker

from rasa.hooks import hookimpl
from rasa.plugin import plugin_manager

INSPECT_TEMPLATE_PATH = "inspector/dist"

structlogger = structlog.get_logger()


class DevelopmentInspectorPlugin:
    """Plugin for broadcasting tracker updates to development inspector clients."""

    def __init__(self, inspector: DevelopmentInspectProxy) -> None:
        self.inspector = inspector
        self.tasks: List[asyncio.Task] = []

    def _cancel_tasks(self) -> None:
        """Cancel all remaining tasks."""
        [task.cancel() for task in self.tasks]
        self.tasks = []

    def _cleanup_completed_tasks(self) -> None:
        """Remove tasks that have already completed."""
        self.tasks = [task for task in self.tasks if not task.done()]

    @hookimpl  # type: ignore[misc]
    def after_new_user_message(self, tracker: DialogueStateTracker) -> None:
        """Broadcasts tracker updates after a new user message."""
        task = asyncio.create_task(self.inspector.on_tracker_updated(tracker))
        self.tasks.append(task)
        self._cleanup_completed_tasks()

    @hookimpl  # type: ignore[misc]
    def after_action_executed(self, tracker: DialogueStateTracker) -> None:
        """Broadcasts tracker updates after an action is executed."""
        task = asyncio.create_task(self.inspector.on_tracker_updated(tracker))
        self.tasks.append(task)
        self._cleanup_completed_tasks()

    @hookimpl  # type: ignore[misc]
    def after_server_stop(self) -> None:
        """Cancels all remaining tasks when the server stops."""
        self._cancel_tasks()


class DevelopmentInspectProxy(InputChannel):
    """Development inspector to inspect channel communication.

    It wraps a Rasa Pro input / output providing an inspect ui showing
    the state of the conversation.
    """

    def __init__(self, underlying: InputChannel, is_voice: bool = False) -> None:
        """Initializes the DevelopmentInspectProxy channel."""
        super().__init__()
        self.underlying = underlying
        self.is_voice = is_voice
        self.processor = None
        self.tracker_stream = TrackerStream(get_tracker=self.get_tracker_state)
        # Register the plugin to get tracker updates
        plugin_manager().register(DevelopmentInspectorPlugin(self))

    def name(self) -> Text:  # type: ignore[override]
        """Channel name."""
        return self.underlying.name()

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> "InputChannel":
        raise NotImplementedError(
            "Method `from_credentials` not supported for the inspect proxy."
        )

    def url_prefix(self) -> Text:
        return self.underlying.name()

    def get_output_channel(self) -> Optional["OutputChannel"]:
        return self.underlying.get_output_channel()

    def get_metadata(self, request: Request) -> Optional[Dict[Text, Any]]:
        return self.underlying.get_metadata(request)

    @staticmethod
    def inspect_html_path() -> Text:
        """Returns the path to the inspect.html file."""
        import pkg_resources

        return pkg_resources.resource_filename(__name__, INSPECT_TEMPLATE_PATH)

    async def get_tracker_state(self, sender_id: str) -> str:
        """Returns the state of the tracker as a json string."""
        if not self.processor:
            structlogger.error(
                "development_inspector.get_tracker_state.agent_not_initialized"
            )
            return ""

        tracker = await self.processor.get_tracker(sender_id)
        state = tracker.current_state(EventVerbosity.AFTER_RESTART)
        return orjson.dumps(state, option=orjson.OPT_SERIALIZE_NUMPY).decode("utf-8")

    async def on_tracker_updated(self, tracker: DialogueStateTracker) -> None:
        """Notifies all clients about tracker updates in real-time."""
        if self.tracker_stream and tracker.sender_id:
            state = tracker.current_state(EventVerbosity.AFTER_RESTART)
            tracker_dump = orjson.dumps(
                state, option=orjson.OPT_SERIALIZE_NUMPY
            ).decode("utf-8")
            await self.tracker_stream.broadcast(tracker_dump)

    async def on_message_proxy(
        self,
        on_new_message: Callable[["UserMessage"], Awaitable[Any]],
        message: "UserMessage",
    ) -> None:
        """Proxies the on_new_message call to the underlying channel."""
        await on_new_message(message)

    @classmethod
    async def serve_inspect_html(cls) -> HTTPResponse:
        """Serves the inspect.html file."""
        return await response.file(cls.inspect_html_path() + "/index.html")

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> "Blueprint":
        """Defines a Sanic blueprint."""
        self.sio = AsyncServer(async_mode="sanic", cors_allowed_origins=[])
        underlying_webhook: Blueprint = self.underlying.blueprint(
            partial(self.on_message_proxy, on_new_message)
        )
        underlying_webhook.static("/assets", self.inspect_html_path() + "/assets")

        @underlying_webhook.route("/inspect.html", methods=["GET"], name="inspect")
        async def inspect(_: Request) -> HTTPResponse:
            return await self.serve_inspect_html()

        @underlying_webhook.listener("after_server_start")  # type: ignore[misc]
        async def after_server_start(app: Sanic, _: asyncio.AbstractEventLoop) -> None:
            """Prints a message after the server has started with inspect URL."""
            self.processor = app.ctx.agent.processor

            inspect_path = app.url_for(f"{app.name}.{underlying_webhook.name}.inspect")

            # replace 0.0.0.0 with localhost
            serve_location = app.serve_location.replace("0.0.0.0", "localhost")

            print_info(
                f"Development inspector for channel {self.name()} is running. To "
                f"inspect conversations, visit {serve_location}{inspect_path}"
            )

        underlying_webhook.add_websocket_route(
            self.tracker_stream, "/tracker_stream", name="tracker_stream"
        )

        return underlying_webhook


class TrackerStream:
    """Stream tracker state to connected clients."""

    def __init__(self, get_tracker: Callable[[str], Awaitable[Dict[str, Any]]]) -> None:
        """Initializes the TrackerStream."""
        self._connected_clients: Set[Websocket] = set()
        self.get_tracker = get_tracker

    def __name__(self) -> str:
        """Name of the stream."""
        return "tracker_stream"

    async def __call__(self, *args: Any, **kwargs: Any) -> None:
        """Starts the stream."""
        await self.stream(*args, **kwargs)

    async def stream(self, request: Request, ws: Websocket) -> None:
        """Handles connection of a new client."""
        self._connected_clients.add(ws)
        try:
            async for message_str in ws:
                message = orjson.loads(message_str)
                # allows frontend to request the tracker state
                # used when websocket begins
                # also used when URL changes (sender updated)
                if message.get("action") == "retrieve":
                    sender_id = message.get("sender_id")
                    if not sender_id:
                        structlogger.warning(
                            "development_insector.tracker_stream.missing_sender_id"
                        )
                        continue
                    tracker_dump = await self.get_tracker(sender_id)
                    await self._send(ws, tracker_dump)
                else:
                    structlogger.warning(
                        "development_inspector.tracker_stream.unknown_action",
                        message=message,
                    )
        finally:
            self._connected_clients.remove(ws)

    async def _send(self, ws: Websocket, message: str) -> None:
        """Sends a message to a connected client."""
        try:
            await ws.send(message)
        except exceptions.WebsocketClosed:
            pass

    async def broadcast(self, message: str) -> None:
        """Broadcasts a message to all connected clients."""
        if not self._connected_clients:
            return
        await asyncio.wait(
            [
                asyncio.create_task(self._send(websocket, message))
                for websocket in self._connected_clients
            ]
        )



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Text

from rasa.core.policies.policy import PolicyPrediction
from rasa.engine.graph import GraphComponent
from rasa.engine.runner.interface import ExecutionContext
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.core.constants import ACTION_LISTEN_NAME
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import (
    ActionExecuted,
    ActionExecutionRejected,
    DefinePrevUserUtteredFeaturization,
)
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import InvalidConfigException, RasaException

logger = logging.getLogger(__name__)


def is_not_in_training_data(
    policy_name: Optional[Text], max_confidence: Optional[float] = None
) -> bool:
    """Checks whether the prediction is empty or by a policy which did not memoize data.

    Args:
        policy_name: The name of the policy.
        max_confidence: The max confidence of the policy's prediction.

    Returns:
        `False` if and only if an action was predicted (i.e. `max_confidence` > 0) by
        a `MemoizationPolicy`
    """
    from rasa.core.policies.memoization import (
        AugmentedMemoizationPolicy,
        MemoizationPolicy,
    )
    from rasa.core.policies.rule_policy import RulePolicy

    if not policy_name:
        return True

    memorizing_policies = [
        RulePolicy.__name__,
        MemoizationPolicy.__name__,
        AugmentedMemoizationPolicy.__name__,
    ]
    is_memorized = any(
        policy_name.endswith(f"_{memoizing_policy}")
        for memoizing_policy in memorizing_policies
    )

    # also check if confidence is 0, than it cannot be count as prediction
    return not is_memorized or max_confidence == 0.0


class InvalidPolicyEnsembleConfig(RasaException):
    """Exception that can be raised when the policy ensemble is not valid."""


class PolicyPredictionEnsemble(ABC):
    """Interface for any policy prediction ensemble.

    Given a list of predictions from policies, which include some meta data about the
    policies themselves, an "ensemble" decides what the final prediction should be, in
    the following way:
    1. If the previously predicted action was rejected, then the ensemble sets the
       probability for this action to 0.0 (in all given predictions).
    2. It combines the information from the single predictions, which include some
       meta data about the policies (e.g. priority), into a final prediction.
    3. If the sequence of events given at the time of prediction ends with a user
       utterance, then the ensemble adds a special event to the event-list included in
       the final prediction that indicates whether the final prediction was made based
       on the actual text of that user utterance.

    Observe that policies predict "mandatory" as well as "optional"
    events. The ensemble decides which of the optional events should
    be passed on.
    """

    def combine_predictions_from_kwargs(
        self, tracker: DialogueStateTracker, domain: Domain, **kwargs: Any
    ) -> PolicyPrediction:
        """Derives a single prediction from predictions given as kwargs.

        Args:
            tracker: dialogue state tracker holding the state of the conversation,
              which may influence the combination of predictions as well
            domain: the common domain
            **kwargs: arbitrary keyword arguments. All policy predictions passed as
              kwargs will be combined.

        Returns:
            a single prediction
        """
        predictions = [
            value for value in kwargs.values() if isinstance(value, PolicyPrediction)
        ]
        return self.combine_predictions(
            predictions=predictions, tracker=tracker, domain=domain
        )

    @abstractmethod
    def combine_predictions(
        self,
        predictions: List[PolicyPrediction],
        tracker: DialogueStateTracker,
        domain: Domain,
    ) -> PolicyPrediction:
        """Derives a single prediction from the given list of predictions.

        Args:
            predictions: a list of policy predictions that include "confidence scores"
              which are non-negative but *do not* necessarily up to 1
            tracker: dialogue state tracker holding the state of the conversation,
              which may influence the combination of predictions as well
            domain: the common domain

        Returns:
            a single prediction
        """
        ...


class DefaultPolicyPredictionEnsemble(PolicyPredictionEnsemble, GraphComponent):
    """An ensemble that picks the "best" prediction and combines events from all.

    The following rules determine which prediction is the "best":
    1. "No user" predictions overrule all other predictions.

    2. End-to-end predictions overrule all other predictions based on
        user input - if and only if *no* "no user" prediction is present in the
        given ensemble.

    3. Given two predictions, if the maximum confidence of one prediction is
        strictly larger than that of the other, then the prediction with the
        strictly larger maximum confidence is considered to be "better".
        The priorities of the policies that made these predictions does not matter.

    4. Given two predictions of policies that are equally confident, the
        prediction of the policy with the higher priority is considered to be
        "better".

    Observe that this comparison is *not* symmetric if the priorities are allowed to
    coincide (i.e. if we cannot distinguish two predictions using 1.-4., then
    the first prediction is considered to be "better").

    The list of events in the final prediction will contain all mandatory
    events contained in the given predictions, the optional events given in the
    "best" prediction, and `DefinePrevUserUtteredFeaturization` event (if the
    prediction was made for a sequence of events ending with a user utterance).
    """

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> DefaultPolicyPredictionEnsemble:
        """Creates a new instance (see parent class for full docstring)."""
        return cls()

    def __str__(self) -> Text:
        return f"{self.__class__.__name__}()"

    @staticmethod
    def _pick_best_policy(predictions: List[PolicyPrediction]) -> PolicyPrediction:
        """Picks the best policy prediction based on probabilities and policy priority.

        Args:
            predictions: a list containing policy predictions

        Returns:
            The index of the best prediction
        """
        best_confidence = (-1.0, -1)
        best_index = -1

        # different type of predictions have different priorities
        # No user predictions overrule all other predictions.
        is_no_user_prediction = any(
            prediction.is_no_user_prediction for prediction in predictions
        )
        # End-to-end predictions overrule all other predictions based on user input.
        is_end_to_end_prediction = any(
            prediction.is_end_to_end_prediction for prediction in predictions
        )

        policy_events = []
        for idx, prediction in enumerate(predictions):
            policy_events += prediction.events

            # No user predictions (e.g. happy path loop predictions)
            # overrule all other predictions.
            if prediction.is_no_user_prediction != is_no_user_prediction:
                continue

            # End-to-end predictions overrule all other predictions based on user input.
            if (
                not is_no_user_prediction
                and prediction.is_end_to_end_prediction != is_end_to_end_prediction
            ):
                continue

            confidence = (prediction.max_confidence, prediction.policy_priority)
            if confidence > best_confidence:
                # pick the best policy
                best_confidence = confidence
                best_index = idx

        if best_index < 0:
            raise InvalidConfigException(
                "No best prediction found. Please check your model configuration."
            )

        best_prediction = predictions[best_index]
        policy_events += best_prediction.optional_events

        return PolicyPrediction(
            best_prediction.probabilities,
            best_prediction.policy_name,
            best_prediction.policy_priority,
            policy_events,
            is_end_to_end_prediction=best_prediction.is_end_to_end_prediction,
            is_no_user_prediction=best_prediction.is_no_user_prediction,
            diagnostic_data=best_prediction.diagnostic_data,
            hide_rule_turn=best_prediction.hide_rule_turn,
            action_metadata=best_prediction.action_metadata,
        )

    @staticmethod
    def _best_policy_prediction(
        predictions: List[PolicyPrediction],
        tracker: DialogueStateTracker,
        domain: Domain,
    ) -> PolicyPrediction:
        """Finds the best policy prediction.

        Args:
            predictions: a list of policy predictions that include "confidence scores"
              which are non-negative but *do not* necessarily up to 1
            tracker: dialogue state tracker holding the state of the conversation,
              which may influence the combination of predictions as well
            domain: the common domain

        Returns:
            The winning policy prediction.
        """
        last_action_event = next(
            (
                event
                for event in reversed(tracker.events)
                if isinstance(event, (ActionExecutionRejected, ActionExecuted))
            ),
            None,
        )

        rejected_action_name = None
        if len(tracker.events) > 0 and isinstance(
            last_action_event, ActionExecutionRejected
        ):
            rejected_action_name = last_action_event.action_name

        if rejected_action_name:
            logger.debug(
                f"Execution of '{rejected_action_name}' was rejected. "
                f"Setting its confidence to 0.0 in all predictions."
            )
            index_of_rejected_action = domain.index_for_action(rejected_action_name)
            for prediction in predictions:
                prediction.probabilities[index_of_rejected_action] = 0.0

        return DefaultPolicyPredictionEnsemble._pick_best_policy(predictions)

    def combine_predictions(
        self,
        predictions: List[PolicyPrediction],
        tracker: DialogueStateTracker,
        domain: Domain,
    ) -> PolicyPrediction:
        """Derives a single prediction from the given list of predictions.

        Note that you might get unexpected results if the priorities are non-unique.
        Moreover, the order of events in the result is determined by the order of the
        predictions passed to this method.

        Args:
            predictions: a list of policy predictions that include "probabilities"
              which are non-negative but *do not* necessarily up to 1
            tracker: dialogue state tracker holding the state of the conversation
            domain: the common domain

        Returns:
            The "best" prediction.
        """
        if not predictions:
            raise InvalidConfigException(
                "Expected at least one prediction. Please check your model "
                "configuration."
            )
        # Reminder: If just a single policy is given, we do *not* just return it because
        # it is expected that the final prediction contains mandatory and optional
        # events in the `events` attribute and no optional events.

        winning_prediction = self._best_policy_prediction(
            predictions=predictions, domain=domain, tracker=tracker
        )

        if tracker.latest_action_name == ACTION_LISTEN_NAME:
            if winning_prediction.is_end_to_end_prediction:
                logger.debug("Made e2e prediction using user text.")
                logger.debug("Added `DefinePrevUserUtteredFeaturization(True)` event.")
                winning_prediction.events.append(
                    DefinePrevUserUtteredFeaturization(True)
                )
            else:
                logger.debug("Made prediction using user intent.")
                logger.debug("Added `DefinePrevUserUtteredFeaturization(False)` event.")
                winning_prediction.events.append(
                    DefinePrevUserUtteredFeaturization(False)
                )

        logger.debug(f"Predicted next action using {winning_prediction.policy_name}.")
        return winning_prediction
